''' Native driver for Starlight Xpress Ultrastar camera.
'''

import time
import array
from concurrent.futures import ThreadPoolExecutor
import numpy as np

from loguru import logger
from kivy.clock import Clock

from jocular.cameras.genericcamera import GenericCamera
from jocular.utils import toast

SX_CLEAR_PIXELS = 1
SX_READ_PIXELS_DELAYED = 2
SX_READ_PIXELS = 3
SX_RESET = 6

# convert int formats
def convert_int(x, from_type='uint16', to_type='uint8'):
	return np.array([x], dtype=from_type).view(dtype=to_type)

class SXUltrastarCamera(GenericCamera):

	def connect(self):
		''' Connect to camera. 
		'''

		if self.connected:
			return

		self.connected = False
		self.sxcamera = None

		try:
			import usb.core
			import usb.util
		except:
			self.status = 'no usb.core or usb.util'
			return

		# try to find Ultrastar
		try:
			self.sxcamera = usb.core.find(idVendor=0x1278, idProduct=0x0525)
			if self.sxcamera is not None:
				logger.info('found Ultrastar')
		except Exception as e:
			logger.trace(f'Cannot find Ultrastar ({e})')

		if self.sxcamera is None:
			self.status = 'cannot find Ultrastar'
			logger.info(self.status)
			return

		try:
			self.sxcamera.get_active_configuration()
		except Exception as e:
			logger.exception(f'problem getting active configuration ({e})')
			self.status = 'cannot get active configuration'
			self.clean_up()
			return

		self.connected = True
		self.width, self.height = 1392, 1040

		self.half_height = self.height # for Ultra, don't divide // 2
		self.lswidth, self.mswidth = convert_int(self.width)
		self.lsheight, self.msheight = convert_int(self.half_height)

		self.status = 'SX Ultrastar: {:} x {:} pixels'.format(
			self.width, self.height)

	def disconnect(self):
		if self.connected:
			self.connected = False
			self.status = 'disconnected'

	def stop_capture(self):
		# Cancel any pending reads
		if hasattr(self, 'capture_event'):
			if not self.internal_timing:  
				# don't cancel reading if using internal timing (causes crash on LS)
				self.capture_event.cancel()

	def clean_up(self):
		logger.debug('cleaning up')
		if hasattr(self, 'sxcamera') and self.sxcamera is not None:
			import usb.util
			usb.util.dispose_resources(self.sxcamera)
			logger.debug('Camera: disposed')

	def handle_failure(self, message, exception=None):
		toast('SX camera problem: {:}'.format(message))
		if exception is not None:
			logger.exception('Camera: handle_failure {:} ({:})'.format(message, exception))
		self.connected = False   # to ensure reconnect if that was the issue
		self.last_capture = None
		try:
			self.clean_up()
		except Exception as e:
			logger.exception('Camera: problem disposing {:}'.format(e))
		finally:
			if self.on_failure is not None:
				self.on_failure()

	def capture(self, exposure=None, on_capture=None, on_failure=None,
		binning=None, internal_timing=False, return_image=False, is_bias=False,
		is_faf=False):

		# is_faf not used in this version

		if is_bias:
			self.exposure = .001

		logger.debug('Capturing from SX')
		# store these so they can be used without argument passing in futures etc
		self.on_capture = on_capture
		self.on_failure = on_failure
		self.exposure = exposure
		self.binning = binning
		self.internal_timing = internal_timing
		self.return_image = return_image
		self.last_capture = None

		# this in orig
		# self.connect()

		if not self.connected:
			toast('SX camera is not connected')
			return

		try:
			if internal_timing:
				logger.debug('SX internally-timed exposure {:3f}'.format(exposure))
				# do internal_timed exposure
				self.start_internal_exposure()
				if return_image:
					time.sleep(exposure)
					return self.sxcamera_read()
				else:
					self.capture_event = Clock.schedule_once(
						self.read_internal_exposure, max(0.05, exposure)
					)
			else:
				# externally-timed
				logger.debug('SX external timed exposure {:3f}'.format(exposure))
				self.start_external_exposure()
				if return_image:
					time.sleep(exposure)
					return self.read_external_exposure()
				else:
					self.capture_event = Clock.schedule_once(
						self.read_external_exposure, max(0.05, self.exposure)
					)

		except Exception as e:
			self.handle_failure('capture', e)

	def exposure_command(self, exposure=None):
		# generate write command for sensor
		logger.debug('starting exposure')
		try:
			rowcode = {'odd': 1, 'even': 2, 'all': 3}
			cmd = SX_READ_PIXELS if exposure is None else SX_READ_PIXELS_DELAYED
			params = array.array('B', [0x40, cmd, rowcode['all'], 0, 0, 0, 10, 0, 
				0, 0, 0, 0, self.lswidth, self.mswidth, self.lsheight, self.msheight, 1, 1])
			if exposure is not None:
				# add exposure details to command
				expo = int(exposure * 1000)  #  ms
				exp1, exp2, exp3, exp4 = convert_int(expo, from_type='uint32')
				params = params + array.array('B', [exp1, exp2, exp3, exp4])
			self.sxcamera.write(1, params, 1000)
		except Exception as e:
			self.handle_failure('exposure_command', e)

	def start_internal_exposure(self):
		try:
			self.exposure_command(exposure=self.exposure)
		except Exception as e:
			self.handle_failure('start_internal_exposure', e)

	def start_external_exposure(self):
		try:
			self.sxcamera.ctrl_transfer(0x40, SX_RESET, 0, 0, 0)
			self.sxcamera.ctrl_transfer(0x40, SX_CLEAR_PIXELS, 0, 0, 0)
			logger.debug('started external exposure')
		except Exception as e:
			self.handle_failure('start_external_exposure', e)

	def read_internal_exposure(self, *args):
		logger.debug('reading internal exposure')
		try:
			pool = ThreadPoolExecutor(3)
			future = pool.submit(self.sxcamera_read)  # thread handles read
			future.add_done_callback(self.image_ready)  # when future is done, call image_ready
		except Exception as e:
			self.handle_failure('read_internal_exposure', e)

	def read_external_exposure(self, *args):
		''' read sensor at end of exposure time
		'''
		try:
			self.exposure_command()
			pix = self.sxcamera.read(0x82, self.height * self.width, timeout=10000)
			pix = np.frombuffer(pix, dtype='uint16').reshape(self.half_height, self.width)
			self.last_capture = pix / 2 ** 16
			if self.on_capture is not None:
				self.on_capture()
		except Exception as e:
			self.handle_failure('read_external_exposure', e)

	def sxcamera_read(self):
		''' This is really the 'fast' read used for framing etc as it reads 
			odd rows and duplicates them; not convinced it is necessary any
			more because deinterlacing is fast; we can then get rid of the
			special case for bias above
		'''
		try:
			logger.debug('getting odd pixels')
			pix = self.sxcamera.read(0x82, self.height * self.width, 10000)
			pix = np.frombuffer(pix, dtype='uint16').reshape(self.half_height, self.width)
			return pix / 2 ** 16
		except Exception as e:
			self.handle_failure('sxcamera_read', e)

	def image_ready(self, future):
		self.last_capture = future.result()  # store image read from sensor
		logger.debug('im stats min {:.4f} max {:.4f} mean {:.4f}'.format(
			np.min(self.last_capture), np.max(self.last_capture), np.mean(self.last_capture)))
		if self.on_capture is not None:
			Clock.schedule_once(self.on_capture, 0)

	def get_capture_props(self):
		''' Any specific info for caller
		'''
		return {
			'camera': 'Ultrastar',
			'pixel_width': 6.45,
			'pixel_height': 6.45
			}

	def get_pixel_height(self):
		return 6.45






''' Simple widgets for Jocular
    See other classes in jocular.widgets for more
        specialised icons (levers etc)
'''

import math
from math import radians

from kivy.lang import Builder
from kivy.uix.label import Label
from kivy.properties import (
    NumericProperty,
    ObjectProperty,
    BooleanProperty,
    StringProperty,
)
from kivy.uix.button import Button
from kivy.uix.togglebutton import ToggleButton
from kivy.uix.textinput import TextInput
from kivy.uix.widget import Widget
from kivy.vector import Vector
from kivy.graphics.transformation import Matrix
from kivymd.uix.button import MDIconButton
from kivymd.uix.textfield import MDTextField
from kivy.uix.behaviors import ToggleButtonBehavior
from kivymd.uix.button import MDFlatButton
from kivymd.uix.slider import MDSlider

from jocular.help import TooltipBehavior


Builder.load_string(
    '''

#:import Clipboard kivy.core.clipboard.Clipboard

<Rotatable>:
    canvas.before:
        PushMatrix
        Translate:
            x: -self.width/2
            y: -self.height/2
        Rotate:
            angle: self.rotangle if self.rotated else 0
            origin: self.center
    canvas.after:
        PopMatrix

<JWidget>:
    # canvas.before:
    #     Color:
    #         rgb: 1, 0, 0
    #         a: .9
    #     Ellipse:
    #         pos: self.pos
    #         size: self.size
    size_hint: None, None
    highlight_color: app.theme_cls.accent_palette
    background_color: .6, 0, .6, 0
    color: app.lowlight_color
    disabled_color: .1, .1, .1, 1   
    markup: True
    halign: 'center'
    valign: 'middle'
    size: self.texture_size
    font_size: app.ring_font_size
    text_size: None, None  # forces size to be that of text
    padding: 2, 2


<JIconButton>:
    text: self.myicon


<JToggleButton>:
    color: app.theme_cls.accent_color if self.state == 'down' else app.lowlight_color


<LabelR>:
    halign: 'right'
    valign: 'middle'
    text_size: self.size
    padding: [dp(7), 0]


<LabelL>:
    halign: 'left'
    valign: 'middle'
    text_size: self.size
    padding: [dp(7), 0]


# vertically aligned text box
<TextInputC>:
    padding: [5, self.height / 2.0 - (self.line_height / 2.0) * len(self._lines), 5, self.height / 2.0 - (self.line_height / 2.0) * len(self._lines)]


# this works with the kivymd master
<JTextField>:
    size_hint: (None, None)
    width: '140dp'
    line_color_normal: 0, 0, 0, 0
    foreground_color: app.theme_cls.accent_color
    text_color_normal: app.theme_cls.accent_color
    font_size: app.form_font_size # '20sp'
    on_text: root.invalid = False
    # for latest kivymd these two lines produce coloured text
    fill_color_normal: .1, 0, 0, 0
    fill_color_focus: .1, 0, 0, .1
    hint_text_color_normal: [1, 0, 0, 1] if (root.invalid and self.text.strip()) else app.hint_color
    mode: 'fill'
    spacing: dp(2)


# I can't get KivyMD toggle behavior to work so this is my implementation
<JMDToggleButton>:
    size_hint: 1, None 
    shift_y: dp(40)
    height: dp(16)
    #width: dp(50)
    md_bg_color: .25, .25, .25, 1
    theme_text_color: 'Custom'
    text_color: .6, .6, .6, 1
    #font_size: '12sp'
    #border: dp(2), dp(2), dp(2), dp(2)
    #padding_x: dp(1)


<Pin>:
    size_hint: None, None
    size: dp(100), dp(20)
    markup: True
    text_color: app.lowlight_color
    theme_text_color: 'Custom'
    background_color: 0, 0, 0, 0

''')


class Pin(MDFlatButton, TooltipBehavior):
    
    def __init__(self, loc=None, comp=None, field='show', show_text='show', 
        tooltip_text='', **kwargs):
        self.comp = comp
        self.field = field
        super().__init__(**kwargs)
        if loc == 'upper-right':
            self.pos_hint = {'right': 1, 'top': 1}
        elif loc == 'upper-left':
            self.pos_hint = {'left': 0, 'top': 1}
        elif loc == 'lower-left':
            self.pos_hint = {'left': 0, 'bottom': 0}
        elif loc == 'lower-right':
            self.pos_hint = {'right': 1, 'bottom': 0}
        if 'left' in loc:
            self.open_text = f'[font=Jocular][size=18sp]>[/size][/font] {show_text}'
            self.close_text = '[font=Jocular][size=18sp]<[/size][/font]'
            self.halign = 'left'
        else:
            self.open_text = f'{show_text} [font=Jocular][size=18sp]<[/size][/font]'
            self.close_text = '[font=Jocular][size=18sp]>[/size][/font]'
            self.halign = 'right'
        self.size_hint = None, None
        self.bind(on_press=self.toggle)
        showing = getattr(self.comp, self.field)
        self.text = self.close_text if showing else self.open_text
        self.tooltip_text = tooltip_text

    def toggle(self, *args):
        cval = getattr(self.comp, self.field)
        setattr(self.comp, self.field, not cval)
        self.text = self.open_text if cval else self.close_text


class JMDToggleButton(MDFlatButton, ToggleButtonBehavior, TooltipBehavior):
    def on_state(self, widget, value):
        widget.text_color = (1, 1, 1, 1) if value == 'down' else (.6, .6, .6, 1)


class JTextField(MDTextField):
    invalid = BooleanProperty(False)


class TextInputC(TextInput):
    pass


class LabelR(Label): 
    pass


class LabelL(Label): 
    pass


class Polar:

    def __init__(self, origin=[0, 0], angle=0, radius=0):
        self.origin = origin
        self.angle = angle
        self.radius = radius

    def to_pos(self):
        ang = self.angle * math.pi / 180
        return (
            self.origin[0] + self.radius * math.cos(ang),
            self.origin[1] + self.radius * math.sin(ang),
        )

    def dist_to(self, pos):
        return Vector(self.to_pos()).distance(pos)


class Rotatable(Widget, Polar):

    rotangle = NumericProperty(0.0)
    rotated = BooleanProperty(True)
    inv = ObjectProperty(Matrix())

    deltas = {
        (1, True): 0,
        (1, False): -90,
        (2, True): -180,
        (2, False): -90,
        (3, True): -180,
        (3, False): 90,
        (4, True): 0,
        (4, False): 90,
    }


    def __init__(self, radial=False, **kwargs):
        super().__init__(**kwargs)
        self.radial = radial
        self.on_angle()


    def on_angle(self, *args):
        q = min(4, (self.angle // 90) + 1)
        self.pos = self.to_pos()
        self.rotangle = self.angle + self.deltas[(q, self.radial)]
        self.generate_inverse()


    def on_width(self, *args):
        self.generate_inverse()


    def on_height(self, *args):
        self.generate_inverse()


    def relocate(self, origin=None, radius=None):
        self.origin = origin
        self.radius = radius
        ang = self.angle * math.pi / 180
        self.x = self.origin[0] + self.radius * math.cos(ang) 
        self.y = self.origin[1] + self.radius * math.sin(ang)
        self.generate_inverse()


    def generate_inverse(self, *args):
        ''' done on width and height as well as init because when 
            using texture size to compute size correct w/h info is 
            not available on init
        '''
        ang = radians(self.rotangle) if self.rotated else 0
        t = Matrix().translate(-self.width / 2 - self.x, -self.height / 2 - self.y, 0)
        t = Matrix().rotate(ang, 0, 0, 1).multiply(t)
        t = Matrix().translate(self.x, self.y, 0).multiply(t)
        self.inv = t.inverse()


    def collide_point(self, x, y):
        ''' transform touchpos using the inverse transform of the rotation;
            takes about 150 microseconds
        '''
        xt, yt, _ = self.inv.transform_point(x, y, 0)
        return (self.x < xt < (self.x + self.width)) and (
            self.y < yt < (self.y + self.height)
        )


class JWidget(Widget, TooltipBehavior):
    pass


class JRotWidget(JWidget, Rotatable):
    pass


class JLabel(JRotWidget, Label):
    pass


class JToggleButton(ToggleButton, JRotWidget):
    pass


class JButton(Button, JRotWidget):
    pass


class JIconTextButton(JButton, JRotWidget):
    pass


class JIconButton(JButton, JRotWidget):
    myicon = StringProperty('')


class JMDIconButton(MDIconButton, Rotatable):
    pass


class JSlider(MDSlider):

    def on_touch_down(self, touch):
        ''' Ensure touch isn't passed thru to scatter
        '''
        if self.collide_point(*touch.pos):
            super().on_touch_down(touch)
            return True
        return False




''' Handles LRGB and L+narrowband processing
'''

import json
from functools import partial

import warnings
warnings.simplefilter('ignore')
import numpy as np

from scipy.stats import trimboth
from skimage.color import rgb2lab, lab2rgb
from skimage.transform import resize, rescale
from loguru import logger

from kivy.properties import NumericProperty, StringProperty
from kivy.app import App
from kivy.metrics import dp
from kivy.uix.gridlayout import GridLayout
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.anchorlayout import AnchorLayout
from kivy.uix.label import Label

from jocular.gradient import estimate_gradient, estimate_background
from jocular.component import Component
from jocular.settingsmanager import JSettings
from jocular.widgets.widgets import JMDToggleButton
from jocular.panel import Panel



def bin_image(im, binfac=2, rescale_to_orig=True):
    ''' bin and rescale to original size; 
        in v0.5 added the preserve range option
    '''
    binfac = int(binfac)
    if binfac == 1:
        return im
    if binfac < 1:
        return im
    im2 = rescale(im, 1 / binfac, anti_aliasing=True, mode='constant', 
        preserve_range=True, multichannel=False)
    if rescale_to_orig:
        return resize(im2, im.shape, anti_aliasing=False, mode='constant',
            preserve_range=True)
    return im2


def _percentile_clip(a, perc=80):
    return np.mean(trimboth(np.sort(a, axis=0), (100 - perc)/100, axis=0), axis=0)


def modify_saturation(x, saturation):
    # Shift endpoints of labA or B to increase saturation. A/B values cover the
    # range -127- to 128 so we increase slope and then trim to range.
    # sat=1 corresponds to shift of 100 (that's a lot!)
    if saturation <= 0:
        return x
    k = saturation * 100
    mx = np.mean(x)
    # steepen but preserve mean (perhaps should do this after trimming?)
    x1 = (x - mx) * (255/(255 - 2*k)) + mx 
    x1[x1 < -127] = -127
    x1[x1 > 127] = 127
    return x1


def modify_hue(x, shift):
    # Shift mean of x by given amount
    x = x + shift
    return x


def limitto01(im):
    im[im < 0] = 0
    im[im > 1] = 1
    return im



class MultiSpectral(Panel, Component, JSettings):
    
    save_settings = [
        'saturation', 
        'colour_stretch', 
        'redgreen', 
        'yellowblue',
        'synth_lum_weights'
        ]

    tab_name = 'Colour'

    configurables = [
        ('bin_colour', {
            'name': 'colour binning', 
            'options': ['1x1', '2x2', '3x3', '4x4'],
            'help': 'Binning the colour channels helps reduce colour noise'}),
        ('compensation', {
            'name': 'compensation', 
            'options': ['none', 'subtract background', 'subtract gradients'],
            'help': 'done prior to colour scaling (subtract gradents recommended)'}),
        ('colour_percentile', {
            'name': 'percentile',
            'float': (99.9, 100.0, .001),
            'help': 'values to include when computing max feature value in colour channel (factory: 99.99)'
            }),
        ('synth_lum_weights', {
            'name': 'syn lum weights',
            'options': ['equal', '30/59/11', '21/72/07'],
            'help': 'weighting of RGB to create synthetic luminance'
            }),
        ('r_weight', {
            'name': 'weight for R channel',
            'float': (.8, 1.2, .01),
            'help': 'multiply R channel by this prior to forming LAB colour'
            }),
        ('g_weight', {
            'name': 'weight for G channel',
            'float': (.8, 1.2, .01),
            'help': 'multiply G channel by this prior to forming LAB colour'
            }),
        ('b_weight', {
            'name': 'weight for b channel',
            'float': (.8, 1.2, .01),
            'help': 'multiply B channel by this prior to forming LAB colour'
            }),
        ]

    spectral_mode = StringProperty('mono')
    luminance_mode = StringProperty('L')
    bin_colour = StringProperty('2x2')
    synth_lum_weights = StringProperty('equal')
    compensation = StringProperty('subtract gradients')
    colour_percentile = NumericProperty(99.99)
    saturation = NumericProperty(0)
    colour_stretch = NumericProperty(.5)
    redgreen = NumericProperty(0)
    yellowblue = NumericProperty(0)
    r_weight = NumericProperty(1)
    g_weight = NumericProperty(1)
    b_weight = NumericProperty(1)


    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.stacker = Component.get('Stacker')
 
        self.app = App.get_running_app()
        try:
            with open(self.app.get_path('multispectral.json'), 'r') as f:
                self.chans = json.load(f)
        except:
            self.chans = {}

        self.lum_chans = ['L', 'synL', 'L+synL', 'H', 'O', 'S']

        self.build()
        self.panel_opacity = 0
        self.set_spectral_mode('mono')

        # these represent what is possible as opposed to what user wants
        self.actual_luminance_mode = 'L'
        self.actual_spectral_mode = 'mono'


    def on_new_object(self):
        self.reset()


    def reset(self):
        # set these to None to force updates in the colour-generation sequence
        self.lum = None
        self.R = None
        self.G = None
        self.B = None
        self.normed_RGB = None
        self.LAB = None          
        self.A_sat = None
        self.B_sat = None
        self.A_hue = None
        self.B_hue = None
        self.RGB = None
        self.layer = None
        self.layer_stretched = None
        self.r_weight = 1
        self.g_weight = 1
        self.b_weight = 1


    def update_info(self):
        if self.actual_luminance_mode:
            self.info(f'{self.actual_luminance_mode}, {self.actual_spectral_mode}')
        else:
            self.info(self.actual_spectral_mode)


    def describe_mode(self):
        if self.actual_spectral_mode == 'mono':
            return 'mono'
        return f'{self.actual_luminance_mode}, {self.actual_spectral_mode}'


    ''' chooser stuff
    '''

    def on_show(self):

        # ensure chooser reflects current state
        filts = self.stacker.get_filters()
        if filts is None:
            filts = {'mono'}

        for c, but in self.chan_buttons.items():
            chan = self.chans[c]
            needs = set(chan['needs'])
            avail = needs <= filts
            but.color = (0, 1, 0, .5) if avail else (1, 0, 0, .3)
            but.disabled = not avail
            but.state = 'down' if c == self.spectral_mode else 'normal'

        for c, but in self.lum_buttons.items():
            if 'L' in c:
                avail = True
            else:
                avail = c in filts
            but.color = (0, 1, 0, .5) if avail else (1, 0, 0, .3)
            but.disabled = not avail
            but.state = 'down' if c == self.luminance_mode else 'normal'


    def _mode_button(self, name):
        return JMDToggleButton(
                text=name, 
                group='channels',
                font_size='16sp',
                #height=dp(12),
                on_press=partial(self.chan_selected, name))


    def _lum_button(self, name):
        return JMDToggleButton(
                text=name, 
                group='lumin',
                font_size='16sp',
                #height=dp(20),
                # width=dp(30),
                #size_hint=(1/9, None),
                on_press=partial(self.lum_selected, name))


    def build(self, *args):

        content = self.contents
        self.chan_buttons = {c: self._mode_button(c) for c in self.chans}
        #anchor_layout = AnchorLayout(anchor_x='center', anchor_y='bottom', size_hint=(1, 1))
        #content.add_widget(anchor_layout)
        #mainbox = BoxLayout(size_hint=(1, None), height=dp(180))

        content.add_widget(Label(text='', size_hint=(1, 1)))

        gl = GridLayout(size_hint=(1, None), cols=6, height=dp(120), spacing=(dp(5), dp(5)))
        for b in self.chan_buttons.values():
            gl.add_widget(b)
        content.add_widget(gl)

        # # spacer
        content.add_widget(Label(text='', size_hint=(1, None), height=dp(15)))

        self.lum_buttons = {c: self._lum_button(c) for c in self.lum_chans}
        bl = GridLayout(size_hint=(1, None), cols=6, height=dp(40), spacing=(dp(5), dp(5)))
        for b in self.lum_buttons.values():
            bl.add_widget(b)
        content.add_widget(bl)

        # anchor_layout.add_widget(mainbox)
        self.app.gui.add_widget(self)


    def chan_selected(self, chan, *args):
        if chan != self.spectral_mode:
            self.set_spectral_mode(chan)
 

    def lum_selected(self, lum, *args):
        if lum != self.luminance_mode:
            self.luminance_mode = lum
            # i.e. do update
            self.set_spectral_mode()


    ''' end of chooser stuff
    '''

    def syn_lum(self, R, G, B):
        if self.synth_lum_weights == 'equal':
            L = (R + G + B) / 3
        elif self.synth_lum_weights == '30/59/11':
            L = .299*R  + .587*G + .114*B
        elif self.synth_lum_weights == '21/72/07':
            L = .2125*R  + .7154*G + .0721*B
        return L / np.percentile(L.ravel(), 99.99)


    def on_saturation(self, *args):
        self.adjust_saturation()


    def on_colour_stretch(self, *args):
        self.adjust_colour_stretch()

    def on_r_weight(self, *args):
        self.create_LAB()

    def on_g_weight(self, *args):
        self.create_LAB()

    def on_b_weight(self, *args):
        self.create_LAB()


    # not currently used
    def on_redgreen(self, *args):
        self.adjust_hue()


    def on_yellowblue(self, *args):
        self.adjust_hue()


    def on_bin_colour(self, *args):
        self.create_LAB()


    def on_subtract_gradients(self, *args):
        self.create_LAB()


    def filters_being_displayed(self):
        ''' return set of filters being used in currently displayed image
        '''
        lum = self.actual_luminance_mode
        spect = self.actual_spectral_mode

        # if mono, we're using all the filters
        if spect == 'mono':
            return {'L', 'R', 'G', 'B', 'H', 'O', 'S'}

        # create set from spect
        f = {s for s in spect.replace('+', '')}

        # check if using 'L'
        if lum in {'L', 'L+synL'}:
            f.add('L')

        # if not using synL, add whatever filter is being used
        elif lum != 'synL':
            f.add(lum)

        return f



    def get_lum_stack(self, R, G, B):
        ''' interpret user's luminance stack preference and record
            what was actually possible in the absence of certain
            channels in the stack
        '''

        if self.luminance_mode == 'L':
            l = self.stacker.get_stack('L')
            if l is None:
                self.actual_luminance_mode = 'synL'
                return self.syn_lum(R, G, B)
            self.actual_luminance_mode = 'L'
            return l

        if self.luminance_mode == 'synL':
            self.actual_luminance_mode = 'synL'
            return self.syn_lum(R, G, B)

        if self.luminance_mode == 'L+synL':
            l = self.stacker.get_stack('L')
            if l is None:
                self.actual_luminance_mode = 'synL'
                return self.syn_lum(R, G, B)
            else:
                self.actual_luminance_mode = 'L+synL'
                return (l + self.syn_lum(R, G, B)) / 2

        # otherwise return channel, or create synthetic lum
        c = self.stacker.get_stack(self.luminance_mode)

        if c is None:
            self.actual_luminance_mode = 'synL'
            return self.syn_lum(R, G, B)

        self.actual_luminance_mode = self.luminance_mode
        return c



    def set_spectral_mode(self, mode=None):
        ''' User selects one of many available spectral
            modes for which there are appropriate
            stacks
        '''

        # might not be updating spectral mode but responding to lum mode change
        if mode is None:
            mode = self.spectral_mode 
        
        self.spectral_mode = mode
        self.actual_spectral_mode = mode

        if not self.stacker.viewing_stack:
            self.app.gui.set('spectral_mode', mode)
            return

        self.reset()
        props = self.chans[mode]

        # get required channels
        stacks = {c: self.stacker.get_stack(c) for c in props['needs']}

        # if any chans not available, do mono
        notavail = np.any([v is None for v in stacks.values()])
        if notavail:
            self.luminance_only()
            self.actual_spectral_mode = 'mono'

        # LRGB-style manipulation
        elif mode in {'RGB', 'HOO', 'SHO', 'HOS'}:
            self.R = stacks[mode[0]]
            self.G = stacks[mode[1]]
            self.B = stacks[mode[2]]
            l = self.get_lum_stack(self.R, self.G, self.B)
            self.lum = Component.get('Monochrome').update_lum(l)
            self.create_LAB()

        # view single filter
        elif mode in {'L', 'H', 'O', 'S', 'R', 'G', 'B'}:
            self.actual_luminance_mode = ''
            Component.get('Monochrome').L_changed(stacks[mode])

        # luminance + narrowband layer
        elif mode.startswith('L+'):
            other = mode[-1]
            L = stacks['L']
            layer = stacks[other]
            self.lum = Component.get('Monochrome').update_lum(L)
            self.layer = layer / np.percentile(layer.ravel(), 99.99)
            self.layer_stretch_changed()
    
        # view in mono
        else:
            self.spectral_mode = 'mono'
            self.luminance_only()

        # update GUI
        self.app.gui.set('spectral_mode', self.spectral_mode)

        self.stacker.update_status()
        self.update_info()


    def stack_changed(self):
        ''' Called by Stacker.stack_changed to handle both mono
            and multispectral cases
        '''

        logger.debug('')

        if self.spectral_mode == 'mono':
            L = self.stacker.get_stack(filt='all')
            Component.get('Monochrome').L_changed(L)
            self.reset()
        else:
            self.set_spectral_mode(self.spectral_mode)         


    ''' these methods are part of the chain of processes used in
        colour manipulation; they generally do some specialised
        computation then call the next process in the chain. Done
        like this for efficiency so only those things that need
        recomputing are
    '''

    def layer_stretch_changed(self):
        if self.layer is not None:
            self.layer_stretched = self.modgamma(self.layer - estimate_gradient(self.layer)[0], 
                g=self.colour_stretch)
            self.layer_sat_changed()


    def layer_sat_changed(self):
        if self.layer_stretched is not None:
            r = 3 * self.saturation * self.layer_stretched + self.lum
            r[r < 0] = 0
            r[r > 1] = 1            
            Component.get('View').display_image(np.stack([r, self.lum, self.lum], axis=-1))


    def luminance_only(self):
        Component.get('Monochrome').L_changed(
            self.stacker.get_stack(filt='all'))


    def luminance_updated(self, lum):
        # called from monochrome if user updates luminance control e.g. white

        self.lum = lum
        # we are in LRGB mode
        if self.LAB is not None:
            self.create_RGB()

        # we are in L+ mode
        elif self.layer is not None:
            self.layer_sat_changed()

        # display as mono
        else:            
            Component.get('View').display_image(lum)


    ''' Colour processes are daisy-chained and intermediate representations cached for speed        
        create_LAB          normed_RGB
        stretch_color       LAB             
        adjust_saturation   A_sat, B_sat
        adjust_hue          A_hue, B_hue
        create_RGB          RGB             <-- entry point for luminosity changes
    '''

    def create_LAB(self):
        # Initial stage of LAB image creation from RGB stacks


        if not self.avail(['R', 'G', 'B']):
            return

        # use lupton 
        # from astropy.visualization import make_lupton_rgb
        # lupton = make_lupton_rgb(self.R, self.G, self.B, Q=10, stretch=0.5)
        # Component.get('View').display_image(lupton / 255)
        # return

        # include color normalisation

        ims = [self.r_weight*self.R, self.g_weight*self.G, self.b_weight*self.B]

        # binning ~ 70 ms
        binfac = int(self.bin_colour[0])
        if binfac > 1:
            ims = [bin_image(im, binfac=binfac) for im in ims]

        # compensation prior to scaling
        if self.compensation == 'subtract gradients':
            ims = [im - estimate_gradient(im) for im in ims]
        elif self.compensation == 'subtract background':
            ims = [im - estimate_background(im)[0] for im in ims]

        ''' the percentile is absolutely critical: setting too low
            leads to saturated stars; too high leads to a washed-out
            colour space
        '''
        # max_pixel_vals = [np.percentile(im.ravel(), 99.99) for im in ims]
        logger.trace('find max')
        if self.colour_percentile > 99.999:
            max_pixel_vals = [np.max(im) for im in ims]
            logger.trace(f'max pixel vals via max {max_pixel_vals}')
        else:
            max_pixel_vals = [np.percentile(im.ravel(), self.colour_percentile) for im in ims]
            logger.trace(f'max pixel vals via percentile {max_pixel_vals}')

        ''' divide all images by the largest of these; this expands the
            colour dynamic range without affected colour ratios; the expanded
            range helps when L is combined
        '''
        ims = [im / max(max_pixel_vals) for im in ims]

        ''' limit image to 0-1 because we are about to stretch
            (while it shouldn't be necessary as we've just divided by
            max feature, there may be some brighter hot pixels 
            above unity)
        '''
        self.normed_RGB = [limitto01(im) for im in ims]

        # rest of process
        self.adjust_colour_stretch()


    def modgamma(self, x, g=.5, a0=.01):
        y = x.copy()
        s = g / (a0 * (g - 1) + a0 ** (1 - g))
        d = (1 / (a0 ** g * (g - 1) + 1)) - 1
        y[x < a0] = x[x < a0] * s
        y[x >= a0] =  (1 + d) * (x[x >= a0] ** g) - d
        return y


    def avail(self, props):
        if type(props) != list:
            props = [props]
        return all([hasattr(self, p) and getattr(self, p) is not None for p in props])


    def adjust_colour_stretch(self, *args):
        # intercept LAB process here if we need to stretch colour

        if self.avail('layer'):
            self.layer_stretch_changed()

        elif self.avail('normed_RGB'):
            # use modified gamma correction here to avoid colour noise
            ims = [self.modgamma(im, g=self.colour_stretch) for im in self.normed_RGB]
            self.LAB = rgb2lab(np.stack(ims, axis=-1))  # 80ms
            self.adjust_saturation()


    def adjust_saturation(self, *args):
        # Modify saturation and call next process
        
        if self.avail('layer'):
            self.layer_sat_changed()

        elif self.avail('LAB'):
            self.A_sat = modify_saturation(self.LAB[:, :, 1], self.saturation)
            self.B_sat = modify_saturation(self.LAB[:, :, 2], self.saturation)
            self.adjust_hue() # next process in hue adjustment


    def adjust_hue(self, *args):
        # Modify hue and generate RGB image
        
        if self.avail('A_sat'):
            self.A_hue = modify_hue(self.A_sat, self.redgreen)
            self.B_hue = modify_hue(self.B_sat, self.yellowblue)
            self.create_RGB()  # next process


    def create_RGB(self):
        ''' Combine L, A and B into RGB image. Called here when previous colour process
        has been applied (hue_changed), or when luminosity component has changed (below)
        '''
        if self.avail(['lum', 'A_hue']):
            self.RGB = lab2rgb(np.stack([ 100 * self.lum, self.A_hue, self.B_hue], axis=-1))
            Component.get('View').display_image(self.RGB)




''' View: represents the actual image on the screen; works with
    ScatterView class
'''

import os
import time
import numpy as np
from loguru import logger

from kivy.app import App
from kivy.graphics.texture import Texture
from kivy.graphics.transformation import Matrix
from kivy.properties import (
    BooleanProperty, NumericProperty, 
    BoundedNumericProperty, ListProperty
    )
from kivy.core.window import Window

from jocular.utils import toast
from jocular.component import Component
from jocular.settingsmanager import JSettings
from jocular.metrics import Metrics
from jocular.widgets.scatterview import ScatterView


class View(Component, JSettings):  # must be in this order

    zoom = NumericProperty(0)
    invert = BooleanProperty(False)
    orientation = BoundedNumericProperty(0, min=0, max=360)
    show_reticle = BooleanProperty(False)
    brightness = NumericProperty(1)
    is_dimmed = BooleanProperty(False)
    transparency = NumericProperty(0)
    ROI = ListProperty(None, allownone=True)
    flip_UD = BooleanProperty(False)
    flip_LR = BooleanProperty(False)
    min_zoom = NumericProperty(.5)
    max_zoom = NumericProperty(30)
    zoom_power = NumericProperty(1)
    continuous_update = BooleanProperty(True)


    configurables = [
        ('flip_UD', {
            'name': 'flip up-down?', 
            'switch': '',
            'help': 'Flip image in the vertical plane. '}),
        ('flip_LR', {
            'name': 'flip left-right?', 
            'switch': '',
            'help': 'Flip image in the horizontal plane'}),
        ('continuous_update', {
            'name': 'update continuously?', 
            'switch': '',
            'help': 'Controls update display continuously by default (takes effect on restart)'}),
        ('min_zoom', {
            'name': 'minimum zoom', 'float': (.2, 1, .1),
            'fmt': '{:.1f} times',
            'help': 'amount of zoom at the low end of the zoom slider'}),
        ('max_zoom', {
            'name': 'max zoom', 'float': (10, 50, 1),
            'fmt': '{:.0f} times',
            'help': 'amount of zoom at the high end of the zoom slider'}),
        ('zoom_power', {
            'name': 'zoom power', 'float': (.5, 2, .1),
            'fmt': 'zoom ^ {:.2f}',
            'help': 'apply a power curve to provide more sensitivity'})
        ]


    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.cached_image = None
        self.last_image_time = 0
        self.app = App.get_running_app()        
        self.scatter = ScatterView(self)
        self.app.gui.add_widget(self.scatter, index=100)
        self.scatter._set_center((0, 0))
        self.ring_selected = False
        self.image_selected = False
        self.old_transparency = self.transparency
        self.old_brightness = self.brightness
        # logger.debug(f'View scatter {self.scatter}')


    def on_new_object(self):
        ''' until we receive the new object size we cannot zoom, 
            so zoom first draw
        '''
        self.reset()


    def on_previous_object(self):
        ''' Reset read for previous observation but also match
            orientation to that stored
        '''
        self.reset()
        self.orientation_settings = Component.get('Metadata').get({'orientation'}) 


    def reset(self):
        # Keep as separate method because most likely called on stack reset
        self.ROI = None
        self.pre_ROI_image = None   # stored so we can undo ROI
        self.last_image = None
        self.reset_texture()
        self.update_state()


    def on_save_object(self):
        Component.get('Metadata').set('orientation', self.orientation)


    def on_show_reticle(self, *args):
        self.app.gui.gui['reticle']['widget'].show = self.show_reticle


    def on_brightness(self, *args):
        self.app.brightness = self.brightness


    def on_transparency(self, *args):
        self.app.transparency = max(0, self.transparency)


    def update_state(self):
        s = 'L-R, ' if self.flip_LR else ''
        s += 'U-D' if self.flip_UD else ''
        if s.endswith(', '):
            s = s[:-2]
        flipstr = f' | flip {s}' if s else ''
        invstr = ' | inv' if self.invert else ''
        self.info(f'rot {self.orientation:.0f}\u00b0 | {self.lever_to_zoom(self.zoom):4.1f}x{flipstr}{invstr}')
        Component.get('Annotator').update()


    def lever_to_zoom(self, z):
        return self.min_zoom + (z ** self.zoom_power) *(self.max_zoom - self.min_zoom)


    def zoom_to_lever(self, z):
        if z < self.min_zoom: 
            return 0
        return ((z - self.min_zoom) / (self.max_zoom - self.min_zoom)) ** (1 / self.zoom_power)


    def on_zoom(self, *args):
        # zoom anchored on window centre
        scale = self.lever_to_zoom(self.zoom) / self.scatter.scale
        mat = Matrix().scale(scale, scale, scale)
        self.scatter.apply_transform(mat, anchor=Metrics.get('origin'))
        self.update_state()


    def on_orientation(self, *args):
        # rotate anchored on window center
        rot = self.orientation - self.scatter._get_rotation()
        self.scatter.apply_transform(Matrix().rotate((np.pi/180)*rot, 0 ,0, 1), anchor=Metrics.get('origin'))
        self.update_state()


    def fit_to_window(self, zero_orientation=True, *args):
        ''' Apply original orientation (from sensor) to image
            and set zoom to just fit to window
        '''
        if self.last_image is None:
            return
        shape = self.last_image.shape
        h, w = shape[0], shape[1]  # done like this because might be 3-D
        if w < 101:
            return
        z = self.zoom_to_lever(Window.height / h)
        App.get_running_app().gui.set('zoom', z)
        self.zoom = z
        if zero_orientation:
            self.orientation = 0
        self.scatter._set_center(Metrics.get('origin'))


    def apply_ROI(self, *args):
        ''' force or undo ROI
        '''

        if Component.get('ObjectIO').existing_object:
            toast('ROI only applies to live captures at present', 3)
            return

        # only allow ROI if we have an image
        if self.last_image is None:
            return

        # if we are undoing a ROI, load pre-ROI image if we have one
        if self.ROI is not None:
            if self.pre_ROI_image is not None:
                self.display_image(self.pre_ROI_image)
                self.fit_to_window(zero_orientation=False)
            # signal ROI change
            self.ROI = None
            return

        ''' before applying a ROI,  store a copy of the current image
        '''
        self.pre_ROI_image = self.cached_image.copy()

        # extract its dims
        shape = self.pre_ROI_image.shape
        h, w = shape[0], shape[1]  # done like this because might be 3-D

        xc, yc = Metrics.get('origin')
        r = Metrics.get('inner_radius') # could add * 1.05 as a margin

        # converts from screen to image pixels
        mapping = self.scatter.to_local

        # pixel coords of centre
        xcp, ycp = mapping(xc, yc)

        # pixel coords of edge
        xep, yep = mapping(xc - r, yc)

        # distance in pixels
        dpix = ((xcp - xep) ** 2 + (ycp - yep) ** 2) ** 0.5

        # don't allow width < min width
        min_width = 32
        dpix = int(max(dpix, min_width // 2))
  
        # ensure image is in eyepiece fully
        if (xcp + dpix > w) or (ycp + dpix > h):
            toast('Must have image completely in eyepiece to apply ROI')
            return

        if self.flip_LR:
            xcp = (w - xcp)
        if self.flip_UD:
            ycp = (h - ycp)

        self.ROI = (
            max(0, int(xcp) - dpix), 
            2 * dpix, 
            max(0, int(ycp) - dpix), 
            2 * dpix)


    def on_ROI(self, *args):
        cam = Component.get('Camera')
        cam.set_ROI(self.ROI)
        self.app.gui.gui['apply_ROI']['widget'].color = \
            self.app.lowlight_color if self.ROI is None else self.app.theme_cls.primary_color

        if self.ROI is not None:
            xstart, width, ystart, height = self.ROI
            im = self.pre_ROI_image[ystart: ystart + height, xstart: xstart + width]
            self.reset_texture(w=width, h=width)
            self.display_image(im)


    def on_invert(self, *args):
        if self.last_image is None:
            return
        # only invert if not colour image    
        if self.scatter.ids.image.texture.colorfmt == 'luminance':
            self.display_image(use_cached_image=True)
            self.update_state()
        elif self.invert:
            self.invert = False


    def on_flip_LR(self, *args):
        self.display_image(use_cached_image=True)
        self.update_state()


    def on_flip_UD(self, *args):
        self.display_image(use_cached_image=True)
        self.update_state()

 
    def reset_texture(self, w=10, h=10, colorfmt='luminance'):
        self.scatter.ids.image.size = w, h
        self.scatter.ids.image.texture = Texture.create(size=(w, h), colorfmt=colorfmt)
        self.scatter._set_center(Metrics.get('origin'))
        self.w, self.h = w, h
        self.colorfmt = colorfmt
        logger.debug(f'texture reset w {w} h {h} colorfmt {colorfmt}')


    def display_blank(self):
        self.display_image(0 * np.ones((self.w, self.h)))


    def do_flips(self, im):
        # perform flips (< 1 ms)
        if self.flip_UD:
            im = im[::-1]
        if self.flip_LR:
            im = np.fliplr(im)
        return im


    def display_image(self, im=None, use_cached_image=False):
        ''' Called with an image, in which case update cached image, perform flips etc
        and display; or without an image, in which case use cached image and perform
        flips/invert directly on that.
        '''

        if (im is None) and (not use_cached_image):
            return

        if use_cached_image and (self.cached_image is None):
            return

        if im is not None:
            self.cached_image = im

        if use_cached_image:
            im = self.cached_image  # just for shorthand below

        # we now have a cached image
        colorfmt = 'luminance' if im.ndim == 2 else 'rgb'

        # check if shape or color has changed
        h, w = im.shape[0], im.shape[1]
        if (w != self.w) or (h != self.h) or (colorfmt != self.colorfmt):
            old_center = self.scatter._get_center()
            self.reset_texture(w=w, h=h, colorfmt=colorfmt)
            self.scatter._set_center(old_center)

        im = self.do_flips(im)

        # only invert if luminance and not RGB image
        if self.invert and colorfmt == 'luminance':
            im = np.subtract(1, im, dtype=im.dtype)

        self.last_image = np.uint8(im * 255)
        self.scatter.ids.image.texture.blit_buffer(
            self.last_image.flatten(), 
            colorfmt=colorfmt, 
            bufferfmt='ubyte')

        # first time through for each object, apply settings
        if hasattr(self, 'orientation_settings') and self.orientation_settings is not None:
            z = self.zoom_to_lever(Window.height / self.last_image.shape[0])
            App.get_running_app().gui.set('zoom', z)
            self.zoom = z
            if 'orientation' in self.orientation_settings:
                self.orientation = self.orientation_settings['orientation']
            self.scatter._set_center(Metrics.get('origin'))
            self.invert = False
            self.orientation_settings = None  # indicates that they have been applied

        # refresh image for monitor at most every 300 ms
        # currently disabled
        if False:
            if (time.time() - self.last_image_time) > .3:
                np.save(os.path.join(self.last_image_dir, f'im_{time.time()}.npy'), 
                    self.last_image)
                self.last_image_time = time.time()




''' Handles all aspects of stack combination, including chooser
    for combination method & stack caching
'''

import numpy as np
from functools import partial
from loguru import logger

from kivy.app import App
from kivy.metrics import dp
from kivy.uix.gridlayout import GridLayout
from kivy.uix.anchorlayout import AnchorLayout
from kivy.properties import StringProperty

from jocular.component import Component
from jocular.widgets.widgets import JMDToggleButton
from jocular.utils import percentile_clip
from jocular.panel import Panel


class StackCombiner(Panel, Component):

    combine_method = StringProperty('mean')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        self.methods = ['mean', '90', '80', '70', 'median']
        self.build()
        self.panel_opacity = 0


    def on_new_object(self):
        self.reset()


    def on_combine_method(self, *args):
        self.app.gui.set('combine_method', self.combine_method)


    def on_show(self):
        for c, but in self.buttons:
            but.state = 'down' if c == self.combine_method else 'normal'


    def _button(self, name):
        return JMDToggleButton(
                text=name, 
                size_hint=(.1, None),
                group='combine',
                font_size='16sp',
                height=dp(20),
                on_press=partial(self.choose, name))


    def build(self, *args):

        content = self.contents
        self.buttons = [(c, self._button(c)) for c in self.methods]
        layout = AnchorLayout(anchor_x='center', anchor_y='bottom', size_hint=(1, 1))
        content.add_widget(layout)
        gl = GridLayout(size_hint=(1, None), cols=5, height=dp(100), spacing=(dp(5), dp(5)))
        for _, b in self.buttons:
            gl.add_widget(b)
        layout.add_widget(gl)
        self.app.gui.add_widget(self)


    def choose(self, method, *args):
        if method != self.combine_method:
            # self.app.gui.set('stretch', stretch)
            self.combine_method = method
            Component.get('Stacker').stack_changed()


    def reset(self):
        self.stack_cache = {}


    def combine(self, stk, orig_sub_map, filt='all', calibration=False):

        logger.trace(f'combining stack for filt {filt}')

        # if calibrating, use user-selected method, otherwise use mean for non-L channels
        if calibration or filt in ['L', 'all']:
            method = self.combine_method
        else:
            method = 'mean' 

        sub_map = {s.name: s for s in stk}
        sub_names = set(sub_map)

        # check if we already have this in the cache
        if filt in self.stack_cache:
            cached = self.stack_cache[filt]
            cached_subs = cached['sub_names']
            ncache = len(cached_subs)

            # we have exactly this set of subs in the cache
            if (cached['method'] == method) and (cached_subs == sub_names):
                return cached['stack']

            # new: if mean combination and diff in just one sub, we can combine quickly
            if method == 'mean':
                onemore = sub_names - cached_subs
                onefewer = cached_subs - sub_names
                if len(onemore) == 1 or len(onefewer) == 1:
                    if len(onemore) == 1:
                        c_op = 1
                        dsub = list(onemore)[0]
                    else:
                        c_op = -1
                        dsub = list(onefewer)[0]
                    if dsub not in orig_sub_map:
                        logger.error(f'sub to add/delete not in orig_sub_map {dsub}')
                        # drop thru to recombination
                    else:
                        stacked = (cached['stack'] * ncache + c_op * orig_sub_map[dsub].get_image()) / (ncache + c_op)
                        self.stack_cache[filt]['stack'] = stacked
                        self.stack_cache[filt]['sub_names'] = sub_names
                        return stacked

        # if not, we need to update
        stacked = combine_stack(stk, method=method)

        logger.trace('-- full stack recompute!')

        # cache results
        self.stack_cache[filt] = {
            'stack': stacked, 
            'method': method,
            'sub_names': sub_names
            }

        return stacked


def combine_stack(subs, method='mean'):
    stk = np.stack([s.get_image() for s in subs], axis=0)
    if len(stk) == 1:
        s = stk[0]
    elif len(stk) == 2:
        s = np.mean(stk, axis=0)
    elif method == 'mean':
        s = np.mean(stk, axis=0)
    elif method == 'median':
        s = np.median(stk, axis=0)
    else: 
        s = percentile_clip(stk, perc=int(method))
    return s




''' Handles calibration library and calibration of subs.
'''

import os.path
import numpy as np
from loguru import logger

from kivy.app import App
from kivy.properties import (
    BooleanProperty, DictProperty, 
    NumericProperty, StringProperty
    )
from kivy.core.window import Window

from jocular.table import Table
from jocular.utils import make_unique_filename, toast, percentile_clip
from jocular.component import Component
from jocular.settingsmanager import JSettings
from jocular.image import Image, save_image, fits_in_dir
from jocular.exposurechooser import exp_to_str
from jocular.gradient import estimate_background

date_time_format = '%d %b %y %H:%M'


def none_to_empty(x):
    return '' if x is None else x


def subregion(m, s):
    ''' Return x and y limits of master (m) that correspond 
        to sub (s), or None if not possible to use this master
    '''

    if m is None:
        return None

    ''' if camera is unknown or different for master and sub,
        only allow subregion processing if the shape is identical
    '''

    if (m.camera is None) or (s.camera is None) or (m.camera != s.camera):
        if (m.shape[0] != s.shape[0]) or (m.shape[1] != s.shape[1]):
            return None

    # extract region covered by m and s
    mx = 0 if m.ROI_x is None else m.ROI_x
    my = 0 if m.ROI_y is None else m.ROI_y
    sx = 0 if s.ROI_x is None else s.ROI_x
    sy = 0 if s.ROI_y is None else s.ROI_y
    mw = m.shape[0] if m.ROI_w is None else m.ROI_w
    mh = m.shape[1] if m.ROI_h is None else m.ROI_h
    sw = s.shape[0] if s.ROI_w is None else s.ROI_w
    sh = s.shape[1] if s.ROI_h is None else s.ROI_h

    # forms subregion bounds
    x0 = sx - mx
    x1 = x0 + sw
    y0 = sy - my
    y1 = y0 + sh

    # if s fits within m, return bounds
    if x0 >= 0 and y0 >= 0 and x1 <= mw and y1 <= mh:
        return (x0, x1, y0, y1)

    return None


class Calibrator(Component, JSettings):

    save_settings = ['apply_dark', 'apply_flat']

    masters = DictProperty({})
    apply_flat = BooleanProperty(False)
    apply_dark = BooleanProperty(False)
    use_l_filter = BooleanProperty(True)
    remove_hot_pixels = BooleanProperty(True)
    fd_exposure_tol = NumericProperty(.1)
    exposure_tol = NumericProperty(5)
    temperature_tol = NumericProperty(5)
    dark_days_tol = NumericProperty(1)
    flat_calibration = StringProperty('bias')

    tab_name = 'Calibration'

    configurables = [
        ('flat_calibration', {
            'name': 'flat calibration method', 
            'options': ['bias', 'flat-dark', 'constant'],
            'help': 'how to calibrate flat subs when generating a master flat'
            }),
        ('fd_exposure_tol', {
            'name': 'flat-dark exposure tol', 
            'float': (0, 2, .1),
            'fmt': '{:.1f} seconds',
            'help': 'If using flat-darks, how close must exposure be (seconds)'
            }),
        ('use_l_filter', {
            'name': 'use L flat?', 
            'switch': '',
            'help': 'If there is no flat for the given filter, use a L flat if it exists'
            }),
        ('exposure_tol', {
            'name': 'exposure tolerance', 
            'float': (0, 30, 1), 
            'fmt': '{:.0f} seconds',
            'help': 'When selecting a dark, select those within this exposure tolerance'
            }),
        ('temperature_tol', {
            'name': 'temperature tolerance', 
            'float': (0, 40, 1),
            'fmt': '{:.0f} degrees',
            'help': 'When selecting a dark, restrict to those within this temperature tolerance'
            }),
        ('dark_days_tol', {
            'name': 'dark age tolerance', 'float': (0, 300, 1),
            'fmt': '{:.0f} days',
            'help': 'Maximum age of darks to use if no temperature was specified'
            }),
        ('remove_hot_pixels', {
            'name': 'remove hot pixels?', 
            'switch': '',
            'help': 'Remove hot pixels when creating calibration masters'
            })
    ]


    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        self.calibration_dir = self.app.get_path('calibration')

        self.masters = {}   # map from name to FITs Image instance
        self.library = {}   # map from name to calibration table info
 
        ''' construct above dicts from calibration FITs in calibration directory
        '''
        for f in fits_in_dir(self.calibration_dir):
            path = os.path.join(self.calibration_dir, f)
            try:
                s = Image(path)
                if s.is_master:
                    self.add_to_library(s)
            except Exception as e:
                logger.warning(f'Calibrator: unable to parse calibration {f} ({e})')


    def on_new_object(self, *args):
        n_masters = len(self.library)
        if n_masters > 0:
            self.info(f'{n_masters:d} masters')
        else:
            self.info('no masters')


    def add_to_library(self, m):
        ''' called on initialisation and when we save a new master
        '''

        # keys are full names so they can be reliably deleted
        self.masters[m.fullname] = m
        self.library[m.fullname] = {
            'name': m.name,
            'camera': none_to_empty(m.camera),
            'type': m.sub_type,
            'exposure': exp_to_str(m.exposure) if m.exposure is not None else '',
            'temperature': str(m.temperature) if m.temperature is not None else '',
            'gain': none_to_empty(m.gain),
            'offset': none_to_empty(m.offset),
            'bin': none_to_empty(m.binning),
            'calibration_method': none_to_empty(m.calibration_method),
            'filter': m.filter,
            'created': m.create_time.strftime(date_time_format),
            'shape_str': m.shape_str,
            'age': m.age,
            'nsubs': m.nsubs if m.nsubs is not None else 0
        }


    def create_master(self, capture_props=None):
        ''' Called by ObjectIO to save an existing stack capture by 
            Jocular as a calibration master
        '''

        if 'filter' not in capture_props:
            toast('Cannot create master: unknown filter', 5)
            return

        stacker = Component.get('Stacker')
        sub_type = capture_props['sub_type']

        if sub_type == 'flat' and capture_props['calibration_method'] == 'None':
            toast('Cannot create master flat: flat is uncalibrated', 5)
            return

        ''' Get hold of master from stacker, forcing the use of stack
            combination method that the user has chosen
        '''
        master = stacker.get_stack(capture_props['filter'], calibration=True)
        capture_props['nsubs'] = stacker.get_selected_sub_count()

        # ''' Remove hot pixels from calibration frame
        # '''


        ''' Flats were divided thru by their robust mean to account for 
            level differences but then scaled to 50% to enable B/W controls; 
            Here, normalise to unity based on central region
        '''
        if sub_type == 'flat':
            w, h = master.shape
            w1, w2 = int(w / 3), int(2 * w / 3)
            h1, h2 = int(h / 3), int(2 * h / 3)
            imr = master[h1: h2, w1: w2]
            robust_mean = percentile_clip(imr.ravel(), perc=75)
            master = master / robust_mean

        self.save_master(data=master, capture_props=capture_props)


    def save_master(self, data=None, capture_props=None):
        ''' Save master and add to library to make it available immediately. Called both by
            create_master above and by the Watched camera for any alien master subs. The difference is
            that create_master above does BPM/flat handling etc so only applies to natively-captured
            calibration masters.
        '''

        name = f'master{capture_props["sub_type"]}.fit'
        path = make_unique_filename(os.path.join(self.calibration_dir, name))

        if self.remove_hot_pixels:
            data = Component.get('BadPixelMap').remove_hot_pixels(data)

        save_image(data=data, path=path, capture_props=capture_props)
        self.add_to_library(Image(path))

        # add to notes field of current DSO
        notes = f'Exposure {exp_to_str(capture_props.get("exposure", 0))}\n'
        notes += '\n'.join([f'{k} {v}' for k, v in capture_props.items() 
            if k in {'filter', 'temperature', 'gain', 'offset', 'camera', 
                'binning', 'calibration_method'}])
        Component.get('Notes').notes = notes

        logger.info(f'new master {capture_props}')


    def calibrate(self, sub):
        ''' Given a light sub, apply calibration. Fails silently if no suitable 
            calibration masters. Note that all flat masters are assumed to be
            calibrated (ie bias or flat-dark has been subtracted)
        '''

        sub.calibrations = set({})

        if len(self.library) == 0:
            self.info('no masters')
            return

        if not (self.apply_dark or self.apply_flat):
            self.info('none')
            return

        logger.trace('starting calibration')

        # get all masters anyway (~1ms)
        dark = self.get_dark(sub)
        flat = self.get_flat(sub)
        bias = self.get_bias(sub)
        
        logger.trace(f'D {dark} F {flat} B {bias}')

        D = self.get_master(dark)
        F = self.get_master(flat)
        B = self.get_master(bias)

        im = sub.get_image()

        # to apply dark we just need a dark
        if self.apply_dark and dark is not None:
            dx0, dx1, dy0, dy1 = subregion(self.masters[dark], sub)
            logger.trace(f'Dark subregion {dx0} {dx1} {dy0} {dy1}')
            im = im - D[dy0: dy1, dx0: dx1]
            sub.calibrations = {'dark'}

        # to apply flat we need a flat and either a bias or a dark
        if self.apply_flat and flat is not None:
            fx0, fx1, fy0, fy1 = subregion(self.masters[flat], sub)
            if 'dark' in sub.calibrations:
                im = im / F[fy0: fy1, fx0: fx1]
                sub.calibrations = {'dark', 'flat'}
            elif bias is not None:
                bx0, bx1, by0, by1 = subregion(self.masters[bias], sub)
                im = (im - B[by0: by1, bx0: bx1]) / F[fy0: fy1, fx0: fx1]
                sub.calibrations = {'bias', 'flat'}
            # else:
            #     # manufacture a constant dark using background
            #     mean_back, std_back = estimate_background(im)
            #     offset = mean_back - 5 * std_back
            #     im = ((im - offset) / F)  + offset
            #     sub.calibrations = {'syn-dark', 'flat'}

        # restore background to avoid clipping in next step 
        if 'bias' in sub.calibrations:
            im = im + np.mean(B)
        elif 'dark' in sub.calibrations:
            im = im + np.mean(D)

        # limit
        im[im < 0] = 0
        im[im > 1] = 1

        sub.image = im

        logger.trace('calibration complete')

        applied = ' '.join(list(sub.calibrations))
        if applied:
            self.info(applied)
        else:
            self.info('none suitable')


    def get_dark(self, sub, exposure_tol=None):
        ''' find darks with same camera, gain, offset, binning, shape, and
            with an exposure that is within tolerance
            NB for backwards compat, don't enforce camera if sub doesn't have one
        '''

        if sub.exposure is None:
            return None

        if exposure_tol is None:
            exposure_tol = self.exposure_tol

        # diagnostics
        if False:
            for k, v in self.masters.items():
                logger.trace('')
                logger.trace(f'Testing {k}, {v}')
                if v.sub_type != 'dark':
                    logger.trace('not dark')
                if v.binning != sub.binning:
                    logger.trace(f'different binning {v.binning} vs {sub.binning}')
                if subregion(v, sub) is None:
                    logger.trace('incompatible subregion')
                if v.gain != (sub.gain if v.gain is not None else v.gain):
                    logger.trace(f'different gain {v.gain} vs {sub.gain}')
                if v.offset != (sub.offset if v.offset is not None else v.offset):
                    logger.trace(f'different offset {v.offset} vs {sub.offset}')
                if v.camera != (sub.camera if v.camera is not None else v.camera):
                    logger.trace(f'different camera {v.camera} vs {sub.camera}')
                if v.exposure is not None and abs(v.exposure - sub.exposure) > self.exposure_tol:
                    logger.trace(f'exposure outside tolerance {v.exposure} vs {sub.exposure}')

        darks = {k: v for k, v in self.masters.items()
                    if v.sub_type == 'dark' and
                        v.binning == sub.binning and
                        subregion(v, sub) is not None and
                        v.gain == (sub.gain if v.gain is not None else v.gain) and
                        v.offset == (sub.offset if v.offset is not None else v.offset) and
                        # v.camera == (sub.camera if v.camera is not None else v.camera) and
                        v.exposure is not None and
                        abs(v.exposure - sub.exposure) <= self.exposure_tol
                }
                    
        temperature = Component.get('Session').temperature

        if temperature is not None:
            # we know temperature, select those with temperatures and within tolerance
            darks = [k for k, v in darks.items() if 
                v.temperature is not None and abs(v.temperature - temperature) < self.temperature_tol]
        else:
            # find those within date tolerance (set to 1 to get darks in current session)
            darks = [k for k, v in darks.items() if v.age < self.dark_days_tol]

        # if we have darks, return name of first one
        return darks[0] if len(darks) > 0 else None


    def get_flatdark(self, sub):
        ''' simply a dark within flat-dark exposure tolerance
        '''
        darks = self.get_dark(sub, exposure_tol=self.fd_exposure_tol)
        if len(darks) == 0:
            logger.debug('no matching flatdark')
            return None

        # find the one with the closest exposure
        for k in sorted(darks, key=darks.get):
            logger.debug(f'matching flatdark {k}')
            return k


    def get_bias(self, sub):
        ''' find bias with same camera, gain, offset, binning, and shape
            NB for backwards compat, don't enforce camera if sub doesn't have one
        '''

        bias = {k: v.age for k, v in self.masters.items()
                    if v.sub_type == 'bias' and
                        v.binning == sub.binning and 
                        subregion(v, sub) is not None and
                        v.gain == (sub.gain if v.gain is not None else v.gain) and
                        v.offset == (sub.offset if v.offset is not None else v.offset) and
                        v.camera == (sub.camera if v.camera is not None else v.camera)
                }

        return min(bias, key=bias.get) if len(bias) > 0 else None
 

    def get_flat(self, sub):
        ''' find flat with same camera, binning, and shape
            NB for backwards compat, don't enforce props if sub doesn't have them
        '''

        flats = {k: v for k, v in self.masters.items()
                    if v.sub_type == 'flat' and
                        v.binning == sub.binning and 
                        subregion(v, sub) is not None and
                        v.camera == (sub.camera if v.camera is not None else v.camera)
                }

        # flat in required filter
        if sub.filter is not None:
            flats_in_filt = {k: v for k, v in flats.items() if v.filter is not None and v.filter == sub.filter}
        else:
            flats_in_filt = {} 

        # if we have none and can use L or C filter, use these
        if (len(flats_in_filt) == 0) and self.use_l_filter:
            flats_in_filt = {k:v for k, v in flats.items() if v.filter == 'L'}

        # do we have any now? if not, return
        if len(flats_in_filt) == 0:
            return None

        # map from flat to difference between sub create time and master create time
        flats = {k: abs(v.create_time - sub.create_time).days for k,v in flats_in_filt.items()}

        # find closest in date
        for k in sorted(flats, key=flats.get):
            return k


    def get_master(self, name):
        if name is None:
            return None
        # Retrieve image (NB loaded on demand, so effectively a cache)
        return self.masters[name].get_image()


    def _most_subs(self, cands):
        c = {k: cands[k]['nsubs'] for k in cands.keys()}
        return max(c, key=c.get)


    def calibrate_flat(self, sub):
        ''' Perform calibrations on flat which include subtracting 
        bias/flat-dark/constant if possible, then rescaling so the 
        mean intensity is .5 (because outlier rejection methods used 
        to combine flat subs work best with normalised 
        frames due to changing light levels; the value of .5 is so that we can 
        use B & W controls; then normalise by robust mean of central region so that
        when in use this region is unaffected.

        Called by Stacker when flats are added to the stack
        '''

        im = sub.get_image()
        sub.calibration_method = 'None'

        # calibrate using selected method
        if self.flat_calibration == 'flat-dark':
            # look for suitable masters within exposure tolerance
            # if none found, look for bias
            flatdark = self.get_flatdark(sub)
            if flatdark is not None:
                im = im - self.get_master(flatdark)
                sub.calibration_method = 'flat-dark'
            
        elif self.flat_calibration == 'bias':
            bias = self.get_bias(sub)
            if bias is not None:
                im = im - self.get_master(bias)
                sub.calibration_method = 'bias'

        elif self.flat_calibration == 'constant':
            # estimate background and subtract 5 sds to get potential lower bound
            bias = self.get_bias(sub)
            if bias is not None:
                mean_back, std_back = estimate_background(self.get_master(bias))
                im = im - (mean_back - 5 * std_back)
                sub.calibration_method = 'constant'

        # normalise by mean of image in central zone 
        w, h = im.shape
        w1, w2 = int(w / 3), int(2 * w / 3)
        h1, h2 = int(h / 3), int(2 * h / 3)
        imr = im[h1: h2, w1: w2]
        robust_mean = percentile_clip(imr.ravel(), perc=75)
        sub.image = .5 * im / robust_mean


    def build_calibrations(self):
        ''' Contruct table from library
        '''

        return Table(
            size=Window.size,
            data=self.library,
            name='Calibration masters',
            description='Calibration masters',
            cols={
                'name': {'w': 120, 'align': 'left', 'label': 'Name', 'action': self.show_calibration_frame},
                'type': {'w': 60, 'label': 'Type', 'align': 'left'},
                'exposure': {'w': 80, 'label': 'Exposure'},
                'temperature': {'w': 80, 'label': 'Temp. C', 'type': str},
                'gain': {'w': 50, 'label': 'Gain', 'type': int},
                'offset': {'w': 60, 'label': 'Offset', 'type': int},
                'bin': {'w': 45, 'label': 'Bin', 'type': int},
                'calibration_method': {'w': 120, 'label': 'Calib?', 'type': str},
                'filter': {'w': 80, 'label': 'Filter'},
                'created': {'w': 130, 'label': 'Created', 'sort': {'DateFormat': date_time_format}},
                'shape_str': {'w': 110, 'label': 'Size'},
                'age': {'w': 50, 'label': 'Age', 'type': int},
                'nsubs': {'w': 50, 'label': 'Subs', 'type': int},
                'camera': {'w': 1, 'align': 'left', 'label': 'Camera', 'type': str},
                },
            actions={'move to delete dir': self.move_to_delete_folder},
            initial_sort_column='created',
            reverse_initial_sort=True,
            on_hide_method=self.app.table_hiding
            )


    def show_calibration_table(self, *args):
        ''' Called when user clicks 'library' on GUI
        '''

        if not hasattr(self, 'calibration_table'):
            self.calibration_table = self.build_calibrations()
        self.app.showing = 'calibration'

        # check for redraw
        if self.calibration_table not in self.app.gui.children:
            self.app.gui.add_widget(self.calibration_table, index=0)

        self.calibration_table.show()    


    def show_calibration_frame(self, row, col, value):
        self.calibration_table.hide()
        # convert row.key to str (it is numpy.str by default)
        # get image from path
        try:
            path = os.path.join(self.calibration_dir, str(row.key))
            im = Image(path, check_image_data=True)
            if im.sub_type == 'flat':
                im.image /= 2
            Component.get('Monochrome').display_sub(im.image)
        except Exception as e:
            logger.error(f'{e}')


    def move_to_delete_folder(self, *args):
        objio = Component.get('ObjectIO')
        for nm in self.calibration_table.selected:
            if nm in self.library:
                objio.delete_file(os.path.join(self.calibration_dir, nm))
                del self.library[nm]
                del self.masters[nm]
        logger.info(f'deleted {len(self.calibration_table.selected)} calibration masters')
        self.calibration_table.update()



''' Star extraction and estimation of star properties such 
    as FWHM/HFD, flux, background, centroids, etc.
    Currently somewhat experimental
'''

import warnings
import math
import numpy as np
from loguru import logger
from scipy.optimize import curve_fit
from skimage.transform import downscale_local_mean, rescale
from skimage.feature import blob_dog
from astropy.stats import mad_std
from photutils.detection import DAOStarFinder
# from photutils.aperture import aperture_photometry, CircularAperture, ApertureStats


def extract_stars(
    im, 
    star_method='photutils', 
    centroid_method='simple',
    target_stars=30,
    nstars=None,
    radius=8,
    binfac=1,
    reset_threshold=False, # for dog
    fwhm_method='count'):
    ''' Find stellar sources using specified method, aiming to find
        target_stars stars. Apply centroid processing within specified
        pixel radius.

        Returns a dict containing centroids, flux, fwhm
    '''

    im2 = im.copy()

    if binfac > 1:
        im2 = downscale_local_mean(im2, (binfac, binfac))

    if star_method == 'photutils':
        stars = _stars_DAO(im2, nstars=target_stars)

    elif star_method == 'DoG':
        stars = DoGStars.extract(
            im2, 
            reset_threshold=reset_threshold, 
            target_stars=target_stars, 
            nstars=nstars)

    if stars is None:
        return {'nstars': 0}

    # undo binning
    if binfac > 1:
        stars[:, :2] *= binfac

    # compute centroids, better flux and fwhm estimates
    if centroid_method == 'simple':
        stardata = simple_starprops(im, stars[:, :2], radius=radius)
    else:
        stardata = starprops(im, stars[:, :2], radius=radius)

    # stardata is N x 4 array of x, y, flux, fwhm
    # re-sort by flux here since it will have been improved 
    inds = np.argsort(np.array(stardata[:, 2]))[::-1]

    return {
        'nstars': len(inds),
        'xcentroid': stardata[inds, 0],
        'ycentroid': stardata[inds, 1],
        'flux': stardata[inds, 2],
        'fwhm': stardata[inds, 3]
    }


def simple_starprops(im, stars, radius=8):
    ''' Compute accurate locations for stars based on star pixel coords.
        This is the original Jocular method that has worked robustly.
    '''

    r = radius
    grid_x, grid_y = np.meshgrid(np.arange(2 * r + 1), np.arange(2 * r + 1))
    mask = np.sqrt((r - grid_x)**2 + (r - grid_y)**2) <= r

    rows, cols = im.shape
    star_data = np.zeros((len(stars), 4))
    ptr = 0
    for x, y in stars:
        x, y = int(x), int(y)
        if (x >= r) and (y >= r) and (x < (cols - r)) and (y < (rows - r)):
            imfrag = im[(y - r):(y + r + 1), (x - r):(x + r + 1)].copy()

            # mean intensity within masked circle
            mean_intensity = np.mean(imfrag[mask])
    
            # star pixels are in mask and above mean intensity
            candidates = mask & (imfrag > mean_intensity)
    
            if candidates.any():
                # estimate background and subtract from image
                mean_background = np.mean(imfrag[np.logical_not(candidates)])
                imfrag_sub = candidates * (imfrag - mean_background)       
                wim = np.sum(imfrag_sub)
                cx = np.sum(grid_x * imfrag_sub) / wim
                cy = np.sum(grid_y * imfrag_sub) / wim
                hfd = compute_fwhm(imfrag, fwhm_method='count')
                star_data[ptr, :] = [cx + x - r, cy + y - r, np.mean(imfrag_sub), hfd]
                ptr += 1  
                 
    return star_data[:ptr, :]


def starprops(im, stars, radius=8, upfac=5, fwhm_method='count'):
    ''' Highly-experimental approach; needs more checking
    '''

    ptr = 0
    rows, cols = im.shape
    star_data = np.zeros((len(stars), 4))
    r = radius
    for x, y in stars:
        x, y = int(x), int(y)

        if (x >= r) and (y >= r) and (x < (cols - r)) and (y < (rows - r)):
            imfrag = im[(y - r):(y + r + 1), (x - r):(x + r + 1)].copy()

            # upsample
            imfrag = rescale(imfrag, upfac, preserve_range=True)
            h, w = imfrag.shape
            rr = (w - 1) // 2

            # remove background (might want to alter radfac so it doesn't use entire stellar radius)
            background = compute_background(imfrag, radfac=1)
            imfrag -= background

            # compute initial starsize
            ss = int(compute_starsize(imfrag, method='count'))

            ''' Use radius starsize to compute centroid in tighter radius, avoiding 
                contamination by other stars in region. Note that this is equivalent
                to a star size of twice the HFD
            '''

            # ensure size is not larger then available radius
            ss = min(ss, rr)

            # extract central part of image and compute its centroid
            cx, cy = centroid(imfrag[rr - ss: rr + ss + 1, rr - ss: rr + ss + 1])

            # if math.isnan(cx) or math.isnan(cy):
            #     print('nan', x, y, cx, cy)

            # convert centroid coordinates back to upscaled image
            cx += (rr - ss)
            cy += (rr - ss)

            ''' recenter image on new centroid cx, cy, which will typically reduce size of 
                image region; ensure that it is square
            '''

            # bug here: cx/cy can sometimes be None
            cxi, cyi = int(cx), int(cy)

            # find max radius of region 
            minr = min(cxi, w - cxi - 1, cyi, w - cyi - 1)

            # select square region
            im3 = imfrag[cyi - minr: cyi + minr + 1, cxi - minr: cxi + minr + 1].copy()

            # find star size using selected method
            radfac = 1 # optionally, could use a smaller radius
            fwhm = compute_starsize(im3, method=fwhm_method, radfac=radfac)

            # compute flux (here for now though it should be done with SAME radius throughout)
            fluxrad = int(min(minr, 2*fwhm))
            flux = np.mean(im3[circular_mask(2 * minr + 1, radius=fluxrad)])

            star_data[ptr, :] = [(cx / upfac) + x - r, (cy / upfac) + y - r, flux / upfac**2, fwhm / upfac]
            ptr += 1 

    return star_data[:ptr, :]


def compute_fwhm(imfrag, upfac=5, fwhm_method='count'):
    ''' using specified method; uses upsampling by factor upfac
        to get better estimate
    '''

    imfrag2 = rescale(imfrag, upfac, preserve_range=True)
    h, w = imfrag2.shape

    # remove background (might want to alter radfac so it doesn't use entire stellar radius)
    background = compute_background(imfrag2, radfac=1)
    imfrag2 -= background

    # compute HFD
    HFD = int(compute_starsize(imfrag2, method=fwhm_method))

    return HFD / upfac


def _stars_DAO(im, nstars=30, fwhm=4, nsigmas=5):
    ''' Star extraction using DAOStarFinder from photutils; return centroids and
        peak pixel value at centroid
    '''

    imc = im - np.median(im)    # subtract rough background estimate
    bkg_sigma = mad_std(imc)    # median absolute deviation
    # extract a few more than requested as we will remove those close to boundary
    daofind = DAOStarFinder(fwhm=fwhm, threshold=nsigmas * bkg_sigma, 
        brightest=int(nstars*1.1))
    stars = daofind(imc)
    if stars is None:
        return None

    return np.transpose([stars['xcentroid'], stars['ycentroid'], stars['peak']])


class DoGStars:
    ''' Simple class that extracts stars using difference of Gaussian
        blob detection. A class is used to enable the threshold for
        extraction to be stored in between successive calls to extract.
    '''

    threshold = None

    @classmethod
    def extract(cls, 
        im, 
        target_stars=30,    # find threshold to detect this many stars
        nstars=None,        # desired number of stars (can be more than target)
        reduce_by=1,        # reduce threshold to allow in more stars (not needed prob.)
        sort_by_mag=True,   # sort by decreasing brightness
        min_sigma=3,
        max_sigma=5,
        reset_threshold=False  # if True, recompute star threshold
        ):
        ''' Find stars in image
        '''

        if reset_threshold:
            cls.threshold = None

        # if threshold is None, find it
        if cls.threshold is None:
            cls.threshold, converged = \
                cls._best_threshold(im, target_stars=target_stars, 
                    tol=.2, maxits=20)
            logger.debug('threshold {:.4f} converged {:}'.format(cls.threshold, converged))
 
        # use Dog-based blob extraction for stars
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            stars = blob_dog(im, 
                min_sigma=min_sigma, 
                max_sigma=max_sigma, 
                threshold=cls.threshold * reduce_by, 
                overlap=0)[:, [1, 0]]

        if len(stars) == 0:
            return None

        # very approximate flux
        flux = [im[int(y), int(x)] for x, y in stars]

        if sort_by_mag:
            inds = np.argsort(flux)[::-1]
            x = np.array([x for x, y in stars])[inds]
            y = np.array([y for x, y in stars])[inds]
            flux = np.array(flux)[inds]

        # select n brightest
        if nstars is None:
            nstars = len(x)

        return np.transpose([x[:nstars], y[:nstars], flux[:nstars]])


    @classmethod
    def _best_threshold(cls, im, target_stars=30, tol=.2, maxits=20):
        ''' Find threshold that delivers target number of stars.
            Stop when within tol or maxits reached.
        '''

        t = .01 
        converged = False
        lo, hi = 0, 10
        its = 0
        while not converged:
            nstars = len(blob_dog(im, min_sigma=3, max_sigma=5, threshold=t, overlap=0))
            its += 1
            diff = nstars - target_stars
            converged = (abs(diff / target_stars) < tol) or (its > maxits)
            if not converged:
                if diff > 0:
                    lo = t
                    t = (hi + t) / 2
                else:
                    hi = t
                    t = (lo + t) / 2
        return t, its <= maxits


''' materials for future photometry
    
    positions = np.transpose((sources['xcentroid'], sources['ycentroid']))
    apertures = CircularAperture(positions, r=3.)
    from photutils import aperture_photometry
    ap = aperture_photometry(imc, apertures)
    aperstats = ApertureStats(imc, apertures)
    flux = sources['flux']
    flux2 = np.array(aperstats.sum)
    flux3 = np.array(ap['aperture_sum'])
    fwhm = np.array(aperstats.fwhm)
'''



''' All the stuff below here is homegrown star property computation in case
    photutils is not used; needs to be rationalised and HFD stuff separated
    out as an alternative to photutils aperturestats computation of FWHM.
'''


def f_gauss(x, A, x0, sigma):
    return A * np.exp(-(x - x0) ** 2 / (2 * sigma ** 2))


def f_moffat(x, A, mu, sigma, beta):
    return A * (1 + ((x-mu)/sigma)**2) ** -beta


def fwhm_moffat(A, mu, sigma, beta):
    return 2 * sigma * (2**(1/beta)-1) **.5


def circular_mask(side, radius=None):
    centre = (side - 1) // 2
    if radius is None:
        radius = side - centre - 1
    return distance_mask(side) <= radius


def distance_mask(side):
    centre = (side - 1) // 2
    Y, X = np.ogrid[:side, :side]
    return np.sqrt((X - centre)**2 + (Y - centre)**2)


def centroid(im):
    ''' should really check if centroid is within im (ie in range of 0..h etc)
        but this presumably occurs only if im has negative pixels, which now it
        doesn't
    '''
    h, w = im.shape
    im2 = im.copy()
    im2[im2 < 0] = 0
    grid_x, grid_y = np.meshgrid(np.arange(w), np.arange(w))
    imsum = np.sum(im2)
    return np.sum(grid_x * im2) / imsum, np.sum(grid_y * im2) / imsum


def compute_flux(im, cx, cy, fwhm=None, n_widths=3):
    ''' compute stellar flux for all centroids cx, cy based on
        estimated fwhm; don't use upsampling (yet)
    '''

    # width of window to compute fwhm in
    r = int(n_widths * fwhm)
    star_mask = circular_mask(2 * r + 1)
    h, w = im.shape

    flux = []
    for xx, yy in zip(cx, cy):
        x, y = int(xx), int(yy)
        if x >= r and y >= r and x + r < w and y + r < h:
            imfrag = im[y - r: y + r + 1, x - r: x + r + 1].copy()
            flux += [np.sum(imfrag[star_mask])]
        else:
            # TO DO: this actually occurs so work out why
            # print('outside bounds in computing flux -- should not happen')
            flux += [0]
    return np.array(flux)


def compute_starsize(im, method='count', radfac=1):
    ''' compute HFD/FWHM based on chosen method
        Assumes im has already had background removed.
        All methods operate on circular star mask that occupies all (if radfac=1) or
        part (radfac < 1) of the analysis window defined by the shape of im.
    '''

    h, w = im.shape
    radius = int(radfac * (h - 1) // 2)
    star_mask = circular_mask(w, radius=radius)

    if method == 'count':
        ''' method used in ASTAP: counts all pixels above
            half peak value and converts count to radius
        '''
        pmax = np.max(im[star_mask]) 
        above = np.count_nonzero(im[star_mask] > 0.5 * pmax)
        return 2 * (above / np.pi) ** .5

    if method == 'moments':
        ''' tends to overestimate FWHM?
        '''
        dist_mask = distance_mask(w)
        return 2 * np.sum(dist_mask[star_mask] * im[star_mask]) / np.sum(im[star_mask])

    if method == 'cumsum':
        ''' original FHD method that also tends to 
            overestimate FWHM
        '''
        dist_mask = distance_mask(w)
        dists = dist_mask[star_mask].ravel()
        order = np.argsort(dists)
        cs = np.cumsum(im[star_mask][order])
        half_point = len(cs) - len(cs[cs > (.5 * cs[-1])])
        return 2*sorted(dists)[half_point]

    if method in {'Gaussian', 'Moffat'}:
        ''' curve-fitting approaches to estimate FWHM; can
            suffer problems of poor fits; could be improved perhaps
            by using all pixels in star mask although trials using
            their radii haven't improved things
        '''     
        h2 = (h - 1) // 2
        # operates using a column crossing central pixel
        counts = im[:, h2].copy()
        rads = np.arange(-h2, h2+1)
        mean = np.sum(counts * rads) / np.sum(counts)
        sigma = np.sqrt(sum(counts * (rads - mean) ** 2) / sum(counts))
        if method == 'Gaussian':
            try:
                p_gaussian, pcov = curve_fit(f_gauss, rads, counts, p0=[max(counts), mean, sigma], maxfev=10000)
                # can sometimes fail to converge
                if math.isnan(p_gaussian[2]):
                    return -1
                return 2.3548 * p_gaussian[2]
            except:
                return -1
        elif method == 'Moffat':
            try:
                p_moffat, pcov = curve_fit(f_moffat, rads, counts, p0=[max(counts), mean, 1, 1], maxfev=10000)
                return fwhm_moffat(*p_moffat)
            except:
                return -1


def compute_background(im, radfac=1):
    ''' compute background in im after excluding central star
        region which takes up radfac of the radius
    '''
    h, w = im.shape
    radius = int(radfac * (h - 1) // 2)
    star_mask = circular_mask(w, radius=radius)
    bkg_pixels = im[~(star_mask)]
    background = np.median(bkg_pixels)
    return background





''' Filterwheel device family
'''

from functools import partial
from loguru import logger

from kivy.clock import Clock
from kivy.properties import StringProperty
from kivymd.uix.button import MDFlatButton
from kivymd.uix.dialog import MDDialog

from jocular.component import Component
from jocular.device import DeviceFamily, Device
from jocular.ascom import connect_to_ASCOM_device
from jocular.utils import toast


class FilterWheel(Component, DeviceFamily):

    modes = {
        'Single': 'SingleFW',
        'Manual': 'ManualFW', 
        'SX EFW': 'SXFW',
        'ASCOM': 'ASCOMFW'
    }

    default_mode = 'Single'
    family = 'FilterWheel'

    current_filter = StringProperty('L')


    def get_state(self):
        return {
            'current_filter': self.current_filter,
            'filtermap': self.filtermap()}


    def filtermap(self):
        # create map from filtername to position e.g. {'L': 1, 'R': 2}
        if self.device is None:
            return {}
        return {getattr(self.device, f'f{p}'): p for p in range(1, 10)}


    def select_filter(self, name='L', changed_action=None, not_changed_action=None):

        logger.debug(f'trying to change to filter {name}')
        self.changed_action = changed_action
        self.not_changed_action = not_changed_action

        # no change of filter
        if self.current_filter == name:
            logger.debug(f'no need to change filter {name}')
            changed_action()
            return

        # filter not in wheel
        if name not in self.filtermap():
            # if we want dark or we currently have dark, special case
            if name == 'dark' or self.current_filter == 'dark':
                if name == 'dark':
                    title = 'Is scope capped for darks?'
                else:
                    title = 'Is scope uncapped for lights?'
                self.dialog = MDDialog(
                    auto_dismiss=False,
                    text=title,
                    buttons=[
                        MDFlatButton(text="DONE", 
                            text_color=self.app.theme_cls.primary_color,
                            on_press=partial(self.confirmed_fw_changed, name)),
                        MDFlatButton(
                            text="CANCEL", 
                            text_color=self.app.theme_cls.primary_color,
                            on_press=self.confirmed_fw_not_changed)
                    ],
                )
                self.dialog.open()

            else:
                toast(f'Cannot find filter {name}')
                logger.warning(f'Cannot find filter {name} in current filterwheel')
                if not_changed_action is not None:
                    not_changed_action()
        else:
            # if we get here, we can go ahead
            self.change_filter(name)


    def confirmed_fw_changed(self, name, *args):
        self.dialog.dismiss()
        self.fw_changed(name)


    def confirmed_fw_not_changed(self, *args):
        self.dialog.dismiss()
        self.fw_not_changed()


    def change_filter(self, name, *args):
        try:
            self.device.select_position(
                position=self.filtermap()[name], 
                name=name, 
                success_action=partial(self.fw_changed, name),
                failure_action=self.fw_not_changed)
        except Exception as e:
            toast(f'problem moving EFW ({e})')
            if self.not_changed_action is not None:
                self.not_changed_action()
            return


    def fw_changed(self, name, dt=None):
        # change has been done
        self.current_filter = name
        logger.debug(f'Filter changed to {name}')
        if self.changed_action is not None:
            self.changed_action()


    def fw_not_changed(self):
        logger.debug('Filter not changed')
        if self.not_changed_action is not None:
            self.not_changed_action()


    def device_connected(self):
        # called by DeviceFamily when a device connects
        if not hasattr(self, 'device'):
            return False

        if self.device is not None:
            self.device.settings_have_changed()


class GenericFW(Device):

    family = StringProperty('FilterWheel')

    f1 = StringProperty('-')
    f2 = StringProperty('-')
    f3 = StringProperty('-')
    f4 = StringProperty('-')
    f5 = StringProperty('-')
    f6 = StringProperty('-')
    f7 = StringProperty('-')
    f8 = StringProperty('-')
    f9 = StringProperty('-')


    def __init__(self, **args):
        ftypes = Component.get('FilterChooser').get_filter_types()
        self.configurables = [('f' + str(i), {'name': 'position ' + str(i), 'options': ftypes}) for i in range(1, 10)]
        super().__init__(**args)


    def settings_have_changed(self):
        
        # only transmit changes for connected devices
        if not self.connected:
            return
        
        ftypes = Component.get('FilterChooser').get_filter_types()

        # remove duplicates and non-existent filters
        logger.debug('Checking for and removing duplicate and non-existent filters')
        filts = set({})
        for pos in range(1, 10):
            fname = f'f{pos}'
            f = getattr(self, fname)
            if f not in ftypes:
                setattr(self, fname, '-')
            if f in filts:
                setattr(self, fname, '-')
            else:
                filts.add(f)
        Component.get('CaptureScript').filterwheel_changed()


class SingleFW(GenericFW):


    def connect(self):
        self.connected = True
        self.status = 'Single filter connected'


    def select_position(self, position=None, name=None, success_action=None, failure_action=None):
        if success_action is not None:
            success_action()


# class SimulatorFW(GenericFW):

#   def select_position(self, position=None, name=None, success_action=None, failure_action=None):
#       if success_action is not None:
#           success_action()

#   def connect(self):
#       self.connected = True
#       self.status = 'Filterwheel simulator active'


class ManualFW(GenericFW):

    def connect(self):
        self.connected = True
        self.status = 'Manual filterwheel'


    def select_position(self, position=None, name=None, success_action=None, failure_action=None):
        self.dialog = MDDialog(
            auto_dismiss=False,
            text=f'Change filter to {name} in position {position}',
            buttons=[
                MDFlatButton(text="DONE", 
                    text_color=self.app.theme_cls.primary_color,
                    on_press=partial(self.post_dialog, success_action)),
                MDFlatButton(
                    text="CANCEL", 
                    text_color=self.app.theme_cls.primary_color,
                    on_press=partial(self.post_dialog, failure_action))
            ],
        )
        self.dialog.open()


    def post_dialog(self, action, *args):
        ''' close dialog and perform action (i.e. success or failure to change)
        '''
        self.dialog.dismiss()
        if action is not None:
            action()


class SXFW(GenericFW):

    def connect(self):

        if self.connected:
            return

        self.connected = False
        logger.debug('importing HID')
        try:
            import hid
        except Exception as e:
            logger.warning(f'Cannot import HID ({e})')
            self.status = 'Cannot import HID'
            return

        try:
            if hasattr(self, 'fw'):
                if self.fw:
                    logger.debug('closing filterwheel')
                    self.fw.close()
            else:
                self.fw = hid.device()
                logger.debug('got HID device')

        except Exception as e:
            logger.warning(f'Failed to get HID ({e})')
            self.status = 'Failed to get HID device'
            return

        # note that if we open when already open, it fails
        try:
            logger.debug('opening SX EFW')
            self.fw.open(0x1278, 0x0920)
            self.connected = True
            self.status = 'SX filterwheel connected'
            logger.debug('successful')
        except Exception as e:
            logger.warning(f'fw.open failed ({e})')
            self.status = 'fw.open failed'
            self.connected = False


    def select_position(self, position=None, name=None, success_action=None, failure_action=None):
        try:
            # move SX filterwheel
            logger.debug(f'Moving SX EFW to position {position}')
            self.fw.write([position + 128, 0])
            # wait three seconds before informing controller it has been done
            logger.debug(f'success so setting up action {success_action}')
            Clock.schedule_once(success_action, 3)
        except Exception as e:
            logger.debug(f'failed to select position ({e})')
            # controller will handle this further
            if failure_action is not None:
                failure_action()


class ASCOMFW(GenericFW):

    driver = StringProperty(None)


    def disconnect(self):
        logger.debug('closing ASCOM filterwheel')
        self.fw.Connected = False


    def connect(self):

        if self.connected:
            return

        res = connect_to_ASCOM_device(device_type='FilterWheel', driver=self.driver)
        self.connected = res.get('connected', False)
        self.status = res['status']
        if self.connected:
            self.driver = res.get('driver', self.driver)
            self.fw = res['device']
        else:
            if 'exception' in res:
                self.status += " ({res['exception']})"


    def select_position(self, position=None, name=None, success_action=None, failure_action=None):
        try:
            # possible 0-indexed?
            position = position - 1
            # move filterwheel
            logger.debug(f'Moving ASCOM EFW to position {position}')
            self.fw.Position = position
            # wait three seconds before informing controller it has been done
            logger.debug(f'success so setting up action {success_action}')
            Clock.schedule_once(success_action, 3)
        except:
            # controller will handle this
            if failure_action is not None:
                failure_action()




''' Handle info.json in its various instantiations over the different versions
'''

import os
import json
from datetime import datetime
from loguru import logger

from jocular.image import fits_in_dir # , update_fits_header
from jocular.component import Component


def remove_empties(d):
    return {
        k: v
        for k, v in d.items()
        if not ((v == '') or (v is None) or (v == {}) or (v == []))
    }


def get_metadata(path, simple=False):
    # Read metadata from path, constructing if necessary
    #  supports info.json (v1/2 of Jocular) and info3.json (v3)
    #  main difference is that info3 is simpler; will always read info3 in pref if both exist

    v1 = False
    try:
        # if we can find v3 metadata, load it
        with open(os.path.join(path, 'info3.json'), 'r') as f:
            md = json.load(f)
    except:
        try:
            #  if pre v3 metadata exists, load it
            with open(os.path.join(path, 'info.json'), 'r') as f:
                md = json.load(f)
            # convert
            newmd = {}
            for p in [
                'Name',
                'Con',
                'OT',
                'Notes',
                'session_notes',
                'SQM',
                'temperature',
                'rejected',
                'seeing',
                'transparency',
                'telescope',
                'camera',
                'exposure',
                'sub_type',
            ]:
                if p in md:
                    newmd[p] = md[p]
            if 'scope' in md and 'orientation' in md['scope']:
                newmd['orientation'] = md['scope']['orientation']
            md = newmd
            v1 = True
        except:
            # cannot find any metadata, so set to empty
            md = {}

    #  if no name, use name of directory contaiining FITs
    if 'Name' not in md:
        md['Name'] = os.path.basename(path)

    md = remove_empties(md)

    # create an infov3 for speed of later loading all obs
    if v1:
        try:
            with open(os.path.join(path, 'info3.json'), 'w') as f:
                json.dump(md, f, indent=1)
        except:
            pass

    #  compute session date/time and number of subs dynamically
    fits = fits_in_dir(path)
    md['nsubs'] = len(fits)
    if len(fits) > 0:
        md['session'] = datetime.fromtimestamp(os.path.getmtime(fits[0])).strftime(
            '%d %b %y %H:%M'
        )

    # convert SQM to sky_brightness now we are handling NELM
    if 'SQM' in md:
        md['sky_brightness'] = md['SQM']
        del md['SQM']

    # remove seeing and transparency words
    for typ in ['seeing', 'transparency']:
        wd = md.get(typ, '')
        if wd == typ + ' unspecified':
            del md[typ]
        elif wd:
            swords = wd.split()
            if swords[0] == typ:
                md[typ] = ' '.join(swords[1:]).strip()
            elif swords[-1] == typ:
                md[typ] = ' '.join(swords[:-1]).strip()

    # handle some legacy changes
    if 'OT' in md and 'Name' in md:
        nm = md['Name'].upper()
        if md['OT'].upper() == 'G+':
            if nm.startswith('ARP') or nm.startswith('VV') or nm.startswith('AM '):
                md['OT'] = 'PG'
            elif nm[:3] in {'HIC', 'PCG', 'SHK'}:
                md['OT'] = 'CG'

    # often we want metadata for observing list in simple format
    if simple:
        return { 
            'Name': md.get('Name', ''),
            'OT': md.get('OT', ''),
            'Con': md.get('Con', ''),
            'Session': md.get('session', ''),
            'N': md.get('nsubs', 0),
            'Notes': md.get('Notes', '')
            }

    return md

class Metadata(Component):

    def on_new_object(self):
        self.md = {}

    def on_previous_object(self):
        self.load(Component.get('ObjectIO').current_object_dir)

    def load(self, path):
        self.md = get_metadata(path)

    def save(self, path, change_fits_headers=False):
        # Save metadata to path in above format

        self.md = remove_empties(self.md)

        if 'Name' not in self.md:
            self.md['Name'] = os.path.basename(path)

        # we no longer want to save this information
        if 'rejected' in self.md:
            del self.md['rejected']
        if 'nsubs' in self.md:
            del self.md['nsubs']
        if 'session' in self.md:
            del self.md['session']

        try:
            with open(os.path.join(path, 'info3.json'), 'w') as f:
                json.dump(self.md, f, indent=1)
        except Exception as e:
            logger.warning(f'Problem saving info3.json to {path} ({e})')

        # change FITs header if requested
        # if change_fits_headers:
        #     logger.debug('Changing fits headers')
        #     for f in fits_in_dir(path):
        #         update_fits_header(f, 
        #             exposure=self.md.get('exposure', None),
        #             sub_type=self.md.get('sub_type', None),
        #             temperature=self.md.get('temperature', None))

    def set(self, field, value=None):
        # set one or more fields of metadata
        if value is None and type(field) == dict:
            for f, v in field.items():
                self.md[f] = v
        else:
            self.md[field] = value

    def get(self, field, default=None):
        # get one or more fields of metadata
        if type(field) == list or type(field) == set:
            d = {}
            for f in field:
                if f in self.md:
                    d[f] = self.md[f]
            return d
        else:
            return self.md.get(field, default)




''' Main app. Mainly used for marshalling resource provision.
'''

import os
import json
import sys
import platform

from kivymd.app import MDApp
from kivy.metrics import dp
from kivy.config import Config
from loguru import logger

from kivy.properties import ListProperty, NumericProperty, OptionProperty, StringProperty
from kivy.clock import Clock
from kivy.core.window import Window
from kivy.core.text import LabelBase

from jocular import __version__
from jocular.component import Component
from jocular.gui import GUI
from jocular.utils import get_datadir, start_logging
from jocular.appearance import sat_to_hue


class Jocular(MDApp):

    showing = OptionProperty(
        'main', options=['main', 'observing list', 'calibration', 'observations', 'subs', 'singlesub']
    )

    lowlight_color = ListProperty([0.5, 0.5, 0.5, 1])
    hint_color = ListProperty([0.25, 0.25, 0.25, 1])
    lever_color = ListProperty([0.32, 0.32, 0.32, 1])
    background_color = ListProperty([0.06, 0.06, 0.06, 0])  # was 1 transp
    line_color = ListProperty([0.3, 0.3, 0.3, 1])  # used in Obj and Table only

    ring_font_size = StringProperty('14sp')
    # info_font_size = StringProperty('14sp')
    form_font_size = StringProperty('15sp')

    data_dir = StringProperty(None)

    subdirs = {'captures', 'calibration', 'snapshots', 'deleted', 
        'exports', 'catalogues', 'settings', 'logs'}

    brightness = NumericProperty(1)
    transparency = NumericProperty(0)


    def get_dir_in_datadir(self, name):
        path = os.path.join(self.data_dir, name)
        try:
            if not os.path.exists(path):
                logger.debug(f'Creating path {path}')
                os.mkdir(path)
            return os.path.join(self.data_dir, name)
        except Exception as e:
            logger.exception(f'Cannot create path {path} ({e})')
            sys.exit('Cannot create subdirectory of Jocular data directory')


    def get_path(self, name):
        ''' centralised way to access resources
        '''

        # if path, return path to data dir, else return path to resource
        if name in self.subdirs:
            return self.get_dir_in_datadir(name)

        # jocular's own resources
        elif name == 'dsos':
            return os.path.join(self.directory, 'dsos')

        # example captures
        elif name == 'example_captures':
            return os.path.join(self.directory, 'example_captures')

        elif name in {
            'configurables.json', 
            'gui.json', 
            'object_types.json', 
            'filter_properties.json',
            'multispectral.json'}:
            return os.path.join(self.directory, 'resources', name)

        # jocular settings
        elif name.endswith('.json'):
            return os.path.join(self.get_dir_in_datadir('settings'), name)

        # specific files
        elif name == 'libusb':
            return os.path.join(self.directory, 'resources', 'libusb-1.0.dll')
        elif name == 'star_db':
            return os.path.join(self.data_dir, 'platesolving', 'star_tiles.npz')
        elif name == 'dso_db':
            return os.path.join(self.data_dir, 'platesolving', 'dso_tiles')
        elif name == 'ASI':
            # NB these need keeping up to date so perhaps better externalised
            if sys.platform.startswith('linux'):
                asi = os.path.join('linux', 'libASICamera2.so.1.21')
                # asi = 'libASICamera2.so.1.18'
            elif sys.platform.startswith('darwin'):
                asi = os.path.join('mac', 'libASICamera2.dylib.1.21')
            else:
                # detect if 32 or 64 bit windows
                bits, _ = platform.architecture()
                if bits.startswith('64'):
                    asi = os.path.join('win64', 'ASICamera2.dll')
                else:
                    asi = os.path.join('win32', 'ASICamera2.dll')
            return os.path.abspath(os.path.join(self.directory, 'resources', asi))

        # everything else is in jocular's own resources
        else:
            return os.path.join(self.directory, 'resources', name)


    def on_brightness(self, *args):
        for c in ['hint_color', 'lowlight_color', 'line_color', 'lever_color']:
            getattr(self, c)[-1] = self.brightness


    @logger.catch()
    def build(self):

        self.icon = self.get_path('jocular.png')
        self.data_dir = get_datadir()

        if self.data_dir is not None:
            start_logging(self.get_path('logs'))

        self.title = f'Jocular v{__version__}'
        self.theme_cls.theme_style = "Dark"     

        LabelBase.register(name='Jocular', fn_regular=self.get_path('jocular4.ttf'))

        try:
            with open(self.get_path('Appearance.json'), 'r') as f:
                settings = json.load(f)
        except:
            # set up initial values for settings
            settings = {'highlight_color': 'BlueGray',  'colour_saturation': 56}

        # apply settings      
        for p, v in settings.items():
            if p == 'highlight_color':
                self.theme_cls.accent_palette = v
                self.theme_cls.primary_palette = v
            elif p.endswith('_color'):
                lg = v / 100
                setattr(self, p, [lg, lg, lg, 1])
            elif p.endswith('font_size'):
                setattr(self, p, f'{v}sp')
            elif p == 'transparency':
                self.transparency = int(v) / 100
            elif p == 'colour_saturation':
                self.theme_cls.accent_hue = sat_to_hue(v)

        self.gui = GUI()

        # draw GUI
        Clock.schedule_once(self.gui.draw, -1)

        return self.gui


    @logger.catch()
    def on_stop(self, exception=None):
        if exception is None:
            logger.info('normal close down')
        else:
            logger.exception(f'root exception: {exception}')
        # save width and height
        Config.set('graphics', 'width', str(int(Window.width / dp(1))))
        Config.set('graphics', 'height', str(int(Window.height /dp(1))))
        Config.write()
        Component.close()
        self.gui.on_close()
        logger.info('finished close down')


    # reset showing to main when any table is hidden
    def table_hiding(self, *args):
        self.showing = 'main'



def startjocular():

    # remove possibility of exiting with escape key
    Config.set('kivy', 'log_level', 'error') # was debug
    Config.set('kivy', 'exit_on_escape', '0')
    Config.write()

    # start app
    try:
        joc = Jocular()
        joc.run()
    except Exception as e:
        ''' any uncaught problems lead to a normal closedown to ensure 
            devices are disconnected, etc.
        '''
        joc.on_stop(exception=e)
        sys.exit(f'Jocular failed with error {e}')




''' Root widget of the application which handles drawing itself and 
    loading components. Most elements are defined declaratively in 
    gui.json in the resources dir.
'''

import os
import json

from functools import partial
from collections import OrderedDict
from loguru import logger
from pathlib import Path

from kivymd.uix.filemanager import MDFileManager
from kivy.app import App
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.boxlayout import BoxLayout
from kivy.clock import Clock
from kivy.core.window import Window
from kivy.lang import Builder

from jocular.widgets.widgets import (
    JLabel, JToggleButton, 
    JIconButton, JButton
    )
from jocular.widgets.jlever import JLever
from jocular.widgets.icons import get_icon
from jocular.uranography import angle360
from jocular.component import Component
from jocular.metrics import Metrics
from jocular.ring import Ring
from jocular.utils import get_datadir, start_logging, add_if_not_exists


Builder.load_string('''

<Splash>:
    canvas:
        Color:
            rgba: .2, .2, .2, 0
        Rectangle:
            pos: self.pos
            size: self.width, self.height
    padding: 20, 20
    orientation: 'vertical'
    size_hint: None, None
    size: dp(450), dp(450)
    pos_hint: {'center_x': 10, 'center_y': .5}
    message: _message 
    Label:
        text: 'Welcome to J[font=Jocular][size=60]o[/size][/font]cular!'
        font_size: '40sp'
        markup: True
    Label:
        halign: 'center'
        valign: 'middle'
        text_size: self.size
        font_size: '20sp'
        text: 'Before you can get started, Jocular needs a place to store captures, calibration files, observations and the like.'
    Label:
        text: ''
    Button:
        size_hint: 
        text: 'Choose Jocular data directory ...'
        font_size: '18sp'
        on_press: root.choose_dir()
    Label:
        id: _message
        text: ''

''')


class Splash(BoxLayout):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
 
    def choose_dir(self, *args):
        ''' create filemanager and open in user's home directory
        '''
        fm = MDFileManager()
        fm.exit_manager = partial(self.exit_filemanager, fm)
        fm.select_path = partial(self.handle_selection, fm)
        fm.show(str(Path.home()))

    def exit_filemanager(self, fm):
        fm.close()

    def handle_selection(self, fm, datadir):

        # timestamp in datadir to check that it is writeable
        try:
            with open(os.path.join(datadir, '.canwrite'), 'w') as f:
                f.write('ok')
        except:
            self.message.text = 'Cannot write to that directory'
            return

        # store absolute path to datadir in .jocular if possible
        try:
            with open(os.path.join(str(Path.home()), '.jocular'), 'w') as f:
                f.write(datadir)
        except:
            self.message.text = 'Cannot write .jocular to home directory'
            return

        # store jocular data dir in main app
        self.app.data_dir = datadir

        # start logging
        start_logging(self.app.get_path('logs'))

        # close filemanager and splashscreen
        fm.close()
        self.hide()

        # if there isn't already a captures/examples directory
        # create one and move example captures to it
        try:
            captures = os.path.join(datadir, 'captures')
            add_if_not_exists(captures)
            # shutil.move(
            #     self.app.get_path('example_captures'), 
            #     captures)
        except Exception as e:
            logger.warning(f'Problem creating captures directory ({e})')

        # and run next steps
        
        Clock.schedule_once(self.app.gui.load_components, 0)

    def show(self, *args):
        self.pos_hint = {'center_x': .5, 'center_y': .5}       

    def hide(self, *args):
        self.pos_hint = {'center_x': 10, 'center_y': .5}       


def _true(val):
    if type(val) == bool:
        return val
    if type(val) == int:
        return val == 1
    if type(val) == str:
        return val.lower() in ['1', 'true', 'yes']
    return False


class GUI(FloatLayout):


    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.app = App.get_running_app()
        self.disabled_controls = set({})
        self.gui = OrderedDict()
        self.bind(size=self.redraw)

        # dictionary recording all components that have changed
        self.changed_components = {}

        # load gui settings or initialise to empty if any issues
        try:
            with open(self.app.get_path('gui_settings.json'), 'r') as f:
                self.config = json.load(f)
        except:
            self.config = {}


    def draw(self, dt=None):

        # construct interface for the first time; we redraw in the order originally drawn
        orig_gui = OrderedDict()

        # Component.get('View')
        reticle = Component.get('Reticle')

        # handle rings first due to draw order
        origin = Metrics.get('origin')
        radii = Metrics.get('ring_radius')
        thickness = Metrics.get('ring_thickness')

        greys = {'background': .02, 'outer': .2, 'middle': .17, 'inner': .12, 'image': 0} # , 'capture_ring': .2}
        for nm, r in radii.items():
            if nm != 'image':
                orig_gui[nm] = {
                    'control_type': 'Ring',
                    'widget': Ring(pos=origin, radius=r, thickness=thickness[nm], grey=greys[nm],
                        start_angle=0, end_angle=360)
                    } 

        # focus/alignment reticle
        orig_gui['reticle'] = {
            'control_type': 'Reticle',
            'widget': reticle
        }
 
        # load GUI spec and convert so widget names are keys
        with open(self.app.get_path('gui.json'), 'r') as f:
            gs = json.load(f)

        for comp, widgets in gs.items():
            for name, widget_spec in widgets.items():
                orig_gui[name] = widget_spec
                orig_gui[name]['component'] = comp

        self.gui = OrderedDict()
        # set up a variable to use for new elements such as ring annotations which need a unique name
        self.new_element_count = 0
        for name, spec in orig_gui.items():
            self.draw_element(name, spec)
            spec['widget'].disabled = True

        self.constructed = True

        # check if we have a data dir
        try:
            data_dir = get_datadir()
            if data_dir is None:
                self.make_splash()
            else:
                self.app.data_dir = data_dir
                Clock.schedule_once(self.load_components, 0)
                logger.info('GUI drawn')
        except Exception as e:
            logger.exception(e)


    def make_splash(self):
        ''' put up welcome screen and get hold of datadir, then load components
        '''
        Window.size = 1200, 800
        s = Splash()
        self.add_widget(s)
        s.show()


    def load_components(self, dt=None):

        for name, spec in self.gui.items():
            spec['widget'].disabled = False

        logger.info('starting to load components')

        for c in ['Status', 'Appearance', 'Metadata', 'Catalogues', 
            'Notes', 'DSO', 'Session', 'Stacker', 
            'Capture', 'CaptureScript', 'Observations', 'ObservingList', 
            'Monochrome', 'MultiSpectral', 'ObjectIO', 'Aligner',
            'DeviceManager', 'Camera', 'FilterWheel', 
            'ExposureChooser',  'FilterChooser', 
            'SettingsManager', 'Stretcher', 'StackCombiner',
            'BadPixelMap', 'Calibrator', 'Snapshotter', 'PlateSolver', 
            'Annotator', 'Help']:
            Component.get(c)

        # bind status to components
        Component.bind_status()

        logger.info('all components loaded')

        # prepare for new object
        Component.initialise_new_object()
        logger.info('new object initialised')


    def redraw(self, *args):
        ''' Redraw interface by relocating components
        '''

        # if still being constructed, do nothing
        if not hasattr(self, 'constructed'):
            return

        origin = Metrics.get('origin')
        ring_radius = Metrics.get('ring_radius')
        thickness = Metrics.get('ring_thickness')
 
        for name, spec in self.gui.items():
            if 'widget' in spec:
                w = spec['widget']
                if spec['control_type'] == 'Ring' and name in ring_radius:
                    w.relocate(origin=origin, radius=ring_radius[name], thickness=thickness[name])
                elif spec['control_type'] == 'Reticle':
                    w.pos = origin
                    w.radius = ring_radius['image']
                else:
                    w.relocate(origin=origin, radius=Metrics.get(spec['location']))


    def draw_element(self, wname, spec):
        ''' Draw element and add to gui dictionary for later redrawing
        '''

        control_type = spec['control_type']

        if control_type in {'Ring', 'Reticle'}:
            self.add_widget(spec['widget'])
            self.gui[wname] = spec 
            return

        angle = spec['angle']
        location = spec['location']
        radial = spec.get('radial', False)
        tooltip = spec.get('tooltip', '')

        # get initial value from config, else from initial, else make it up

        if spec.get('group', '-') in self.config:
            initial_value = wname == self.config[spec['group']]
        elif wname in self.config:
            initial_value = self.config[wname]
        elif 'initial' in spec:
            initial_value = spec['initial']
        elif 'values' in spec:
            initial_value = spec['values'][0]
        else:
            initial_value = 0

        # ensure that initial value is within range
        if 'values' in spec:
            if initial_value > spec['values'][1]:
                initial_value = spec['values'][1]
                logger.debug(f'resetting initial_value of {wname} to {initial_value}')
            elif initial_value < spec['values'][0]:
                initial_value = spec['values'][0]
                logger.debug(f'resetting initial_value of {wname} to {initial_value}')
            # print(f'checking {initial_value} is in {spec["values"]}')

        rad = Metrics.get(location)
        orig = Metrics.get('origin')
        radial = spec.get('radial', False)

        # handle icons uniformly
        if 'icon' in spec and 'text' in spec:
            text = get_icon(
                spec['icon'], 
                font_size=spec.get('font_size', None)) + '\n' + spec['text']  
        elif 'icon' in spec:
            text = get_icon(
                spec['icon'], 
                font_size=spec.get('font_size', None))
        else:
            text = spec.get('text', '')

        if control_type == 'JToggleButton':
            w = JToggleButton(angle=angle, origin=orig, radius=rad, text=text, radial=radial,
                allow_no_selection=False)
            if 'group' in spec:
                w.group = spec['group']
            w.state = 'down' if _true(initial_value) else 'normal'

        elif control_type == 'JButton':
            w = JButton(angle=angle, origin=orig, radius=rad, radial=radial, 
                text=text)

        # to do: see if we can combine with JButton
        elif control_type == 'JIconButton':
            w = JIconButton(angle=angle, origin=orig, radius=rad, myicon=text)

        elif control_type == 'JLabel':
            w = JLabel(angle=angle, origin=orig, radius=rad, text=spec.get('text', ''))

        elif control_type == 'JLever':

            if spec.get('groove', False):
                min_a, max_a = angle360(90 - angle[0]), angle360(90 - angle[1])
                w = Ring(pos=orig, radius=rad, thickness=Metrics.get('outer_width')/2,
                    grey=.13, start_angle=angle360(min(min_a, max_a)), 
                    end_angle=angle360(max(min_a, max_a)))
                self.gui[f'_new{self.new_element_count}'] = {
                    'control_type': 'Ring', 
                    'widget': w,
                    'save': False,
                    'location': location}
                self.new_element_count += 1
                self.add_widget(w)

            w = JLever(angles=angle, origin=orig, radius=rad, value=initial_value, 
                    values=spec['values'],
                    continuous_update=spec.get('continuous', None),
                    shortcut=spec.get('shortcut', None),
                    sc_inc=spec.get('sc_inc', None),
                    text=spec.get('text', get_icon('lever', font_size=19)),
                    radial=spec.get('radial', True))

            self.add_widget(w)  # added here to get depth order right

            if spec.get('major', False):
                for vstr, v in spec['major'].items():
                    lab = JLabel(
                        radial=spec.get('radial', True), text=vstr, 
                        angle=w.value_to_angle(v), origin=orig, radius=rad)
                    self.add_widget(lab) 
                    self.gui[f'_new{self.new_element_count}'] = {
                        'control_type': 'JLabel', 
                        'widget': lab,
                        'save': False,
                        'location': location}
                    self.new_element_count += 1

        # post inits (may need more)
        if not 'icon' in spec:
            for f in ['font_size', 'color', 'text_width', 'background_color', 'halign', 'radial']:
                if f in spec:
                    setattr(w, f, spec[f])
  
        if control_type == 'JLever':
            w.bind(value=partial(self.on_action, wname))
        else:
            self.add_widget(w)
            w.bind(on_press=partial(self.on_action, wname))

        w.tooltip_text = tooltip
        spec['widget'] = w
        self.gui[wname] = spec


    def initialise_component(self, component):
        ''' Push initial settings into the component when it is loaded
        '''

        for k, spec in self.gui.items():
            if spec.get('component', None) == component:
                # do NOT do actions!
                if 'action' not in spec:
                    if 'group' in spec:
                        if spec['widget'].state == 'down':
                            self.on_action(k)
                    else:
                        self.on_action(k)


    def get_gui_setting(self, key):
        if hasattr(self, 'config'):
            return self.config.get(key, None)
        return None


    def set(self, name, value, update_property=False):
        ''' Called by any component that needs to access GUI element, usually
            for something simple like changing text. Option to pass this
            change back to component via update property
        '''

        if name not in self.gui:
            return
        spec = self.gui[name]
        ctype = spec['control_type']
        if ctype in ['JLabel', 'JButton', 'JIconButton']:
            setattr(spec['widget'], 'text', str(value))
        elif ctype == 'JLever':
            getattr(spec['widget'], 'reset_value')(value)
        elif ctype == 'JToggleButton':
            # if it is a group, set all but this to normal
            if 'group' in spec:
                gname = spec['group']
                for n, v in self.gui.items():
                    if v.get('group','') == gname:
                        v['widget'].state = 'down' if n == name else 'normal'
            else:
                setattr(spec['widget'], 'state', 'down' if value else 'normal')
        if update_property:
            self.on_action(name)


    def set_prop(self, name, prop, value):
        if name not in self.gui:
            return
        spec = self.gui[name]
        if prop == 'text' and spec['control_type'] == 'JLever':
            spec['widget'].text = value
        else:
            if hasattr(spec, prop):
                setattr(spec, prop, value)


    def get(self, name):
        ''' Access value of a given component; occasionally required
        '''

        if name not in self.gui:
            return None
        spec = self.gui[name]
        w = spec['widget']
        ctype = spec['control_type']
        if ctype in ['JLabel', 'JButton']:
            return w.text
        if ctype == 'JLever':
            return w.value
        if ctype == 'JToggleButton':
            return 1 if w.state == 'down' else 0
        return None


    def on_action(self, name, *args):
        ''' Intercept any GUI event in order to load component, before
            passing action on to component via a property change
        '''

        spec = self.gui[name]
        comp = spec['component']
        c = Component.get(comp)  # loads if needed

        if name in self.disabled_controls:
            return

        # if action is specified, simply call that method on component
        if 'action' in spec:
            getattr(c, spec['action'])()
            return

        # otherwise it involves a value change
        control = spec['control_type']

        # handle button groups
        if control == 'JToggleButton':
            if 'group' in spec:
                setattr(c, spec['group'], spec['widget'].text)
                return

        if hasattr(c, name):
            if control == 'JToggleButton':
                setattr(c, name, spec['widget'].state == 'down')
            elif control == 'JLever':
                setattr(c, name, spec['widget'].value)
            # elif control == 'JMulti':
            #     setattr(c, name, spec['widget'].text)


    def enable(self, names=None):
        # None indicates enable everything
        if names is None:
            names = self.disabled_controls
        # remove names from disabled list
        self.disabled_controls = self.disabled_controls - set(names)
        for n in names:
            if n in self.gui:
                spec = self.gui[n]
                spec['widget'].disabled = False


    def disable(self, names):
        # adds names to disabled list
        self.disabled_controls = self.disabled_controls.union(set(names))
        for n in names:
            if n in self.gui:
                spec = self.gui[n]
                spec['widget'].disabled = True


    def on_close(self, *args):
        ''' Save certain GUI values to config. We save everything except
            actions and elements whose save property is False
        '''
        pass # actually done in Component now


    ''' Leaving this here por si acaso
        uses plyer, but that is super-ugly on Windows, but keep code in case
        kivymd filemanager breaks down....
    '''

    # def choose_dir(self, *args):
    #   filechooser.choose_dir(
    #       on_selection=self._handle_selection,
    #       multiple=False)

    # def handle_selection(self, selection):
    #   datadir = os.path.abspath(selection[0])

    #   # timestamp in datadir to check that it is writeable
    #   try:
    #       with open(os.path.join(datadir, '.canwrite'), 'w') as f:
    #           f.write('ok')
    #   except Exception as e:
    #       self.message.text = 'Cannot write to that directory'
    #       return

    #   # store absolute path to datadir in .jocular if possible
    #   try:
    #       with open(os.path.join(str(Path.home()), '.jocular'), 'w') as f:
    #           f.write(datadir)
    #   except Exception as e:
    #       self.message.text = 'Cannot write .jocular to home directory'
    #       return

    #   # store jocular data dir in main app
    #   App.get_running_app().data_dir = datadir

    #   # and close panel
    #   self.hide()

    #   # and run next steps
    #   Clock.schedule_once(self.gui.load_components, 0)




''' Handles DSO details including lookup and catalogue entry
'''

import math
from functools import partial
from loguru import logger

from kivy.app import App
from kivy.properties import StringProperty, BooleanProperty, DictProperty
from kivy.lang import Builder
from kivymd.uix.boxlayout import MDBoxLayout
from kivymd.uix.button import MDFlatButton
from kivymd.uix.dialog import MDDialog

from jocular.component import Component
from jocular.RA_and_Dec import RA, Dec
from jocular.widgets.widgets import Pin

Builder.load_string('''

<DSOBoxLayout@MDBoxLayout>:
    size_hint: (1, None)
    height: '42dp'

<DSO_panel>:
    name_field: _name
    ot_field: _ot
    padding: '2dp' # '10dp'
    adaptive_height: True
    pos_hint: {'top': .99, 'x': 0} if root.dso.show_DSO else {'top': .99, 'right': -1000} 
    size_hint: None, None
    orientation: 'vertical'
    spacing: dp(10)

    BoxLayout:
        size_hint: (1, None)
        height: '48dp'
        JTextField:
            id: _name
            width: '300dp'
            height: '40dp'
            hint_text: 'name'
            on_text: root.dso.Name_changed(self.text)
            font_size: f'{int(app.form_font_size[:-2]) + 8:}sp' # 28sp

    DSOBoxLayout:
        JTextField:
            id: _ot
            hint_text: 'type'
            #helper_text: 'e.g. PN, GX'
            on_focus: root.dso.OT_changed(self) if not self.focus else None
            text: root.dso.OT
        JTextField:
            width: '140dp'
            hint_text: 'con'
            #helper_text: 'e.g. PER'
            on_text: root.dso.prop_changed(self, 'Con')
            text: root.dso.Con

    DSOBoxLayout:
        JTextField:
            hint_text: 'RA'
            #helper_text: "e.g. 21h30'42"
            on_text: root.dso.prop_changed(self, 'RA', update=False)
            text: '' if root.dso.RA == 'nan' else root.dso.RA
            on_focus: root.dso.prop_changed(self, 'RA') if not self.focus else None

        JTextField:
            width: '130dp'
            hint_text: 'dec'
            #helper_text: "-3 21' 4"
            on_text: root.dso.prop_changed(self, 'Dec', update=False)
            text: '' if root.dso.Dec == 'nan' else root.dso.Dec
            on_focus: root.dso.prop_changed(self, 'Dec') if not self.focus else None

    DSOBoxLayout:
        JTextField:
            hint_text: 'diam'
            #helper_text: "e.g. 21'"
            on_text: root.dso.prop_changed(self, 'Diam', update=False)
            text: '' if root.dso.Diam == 'nan' else root.dso.Diam
            on_focus: root.dso.prop_changed(self, 'Diam') if not self.focus else None

        JTextField:
            width: '110dp'
            hint_text: 'mag'
            #helper_text: "e.g. 14.1"
            # on_focus: root.dso.prop_changed(self, 'Mag') if not self.focus else None
            on_text: root.dso.prop_changed(self, 'Mag')
            text: '' if root.dso.Mag == 'nan' else root.dso.Mag

    DSOBoxLayout:
        JTextField:
            width: '200dp'
            hint_text: 'other'
            #helper_text: ""
            text: root.dso.Other
            on_text: root.dso.prop_changed(self, 'Other')

''')


def prop_to_str(prop, val):
    ''' convert DSO to canonical str format for display/comparison purposes
    '''
    if prop in {'Con', 'Name', 'OT', 'Other'}:
        return val
    if prop == 'RA':
        return str(RA(val)) if val else ''
    if prop == 'Dec':
        return str(Dec(val)) if val else ''
    if prop == 'Mag':
        return float_to_str(val)
    if prop == 'Diam':
        return arcmin_to_str(val)


def float_to_str(x):
    ''' nans with return empty string
    '''
    try:
        s = str(x)
        if s == 'nan':
            return ''
        return s
    except:
        return ''


def str_to_float(s):
    try:
        return float(s)
    except:
        return math.nan


def str_to_arcmin(diam):
    ''' convert string representation of diameter to float, taking
        account of possible suffices (degrees, min, secs); 
        need to raise correct exception (to do)
    '''
    diam = diam.strip()
    if diam == 'nan':
        return math.nan
    if not diam:
        return math.nan
    if diam.endswith('d') or diam.endswith('\u00b0'):
        return float(diam[:-1]) * 60
    if diam.endswith('"'):
        return float(diam[:-1]) / 60
    if diam.endswith("'"):
        return float(diam[:-1])
    return float(diam)


def arcmin_to_str(diam):
    ''' convert float representation of diam to a string with
        the most suitable suffix (degree, arcmin, arcsec)
    '''
    if type(diam) == str:
        return diam
    if math.isnan(diam):
        return ''
    if diam > 60:
        return f"{diam / 60:.2f}\u00b0"
    if diam < 1:
        return f'{diam * 60:.2f}"'
    return f"{diam:.2f}\'"


def validated_prop(val, prop):
    ''' is val a valid property of type prop
    '''
    if prop == 'OT':
        return None if len(val) != 2 else val
    if prop == 'Con':
        return None if len(val) != 3 else val
    if prop == 'RA':
        RA_deg = RA.parse(val)        
        return None if RA_deg is None else str(RA(RA_deg))
    if prop == 'Dec':
        Dec_deg = Dec.parse(val)
        return None if Dec_deg is None else str(Dec(Dec_deg))
    if prop == 'Mag':
        try:
            return str(float(val))
        except:
            return None
    if prop == 'Diam':
        try:
            diam_str = str_to_arcmin(val)
            return arcmin_to_str(diam_str)
        except:
            return None
    if prop == 'Other':
        return val
    return False


class DSO_panel(MDBoxLayout):
    ''' visual representation of editable DSO properties
    '''

    def __init__(self, dso, **kwargs):
        self.dso = dso
        super().__init__(**kwargs)


class DSO(Component): # , JSettings):

    save_settings = ['show_DSO']

    Name = StringProperty('')
    Con = StringProperty('')
    RA = StringProperty('')
    Dec = StringProperty('')
    OT = StringProperty('')
    Mag = StringProperty('')
    Diam = StringProperty('')
    Other = StringProperty('')
    otypes = DictProperty({})
    updating = BooleanProperty(False)
    original_props = DictProperty(None, allownone=True)

    props = ['Name', 'Con', 'OT', 'RA', 'Dec', 'Mag', 'Diam', 'Other']

    show_DSO = BooleanProperty(False)


    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        show_DSO = self.app.gui.get_gui_setting('show_DSO')
        self.show_DSO = False if show_DSO is None else show_DSO
        self.cats = Component.get('Catalogues')
        otypes = self.cats.get_object_types()
        self.otypes = {k: v['name'] for k, v in otypes.items()}
        self.dso_panel = DSO_panel(self)
        self.app.gui.add_widget(self.dso_panel)        
        self.app.gui.add_widget(Pin(
            comp=self, 
            field='show_DSO', 
            loc='upper-left', 
            tooltip_text='toggle DSO information panel',
            show_text='DSO'))


    def on_new_object(self):
        ''' Called when user clicks new
        '''
        self.update_properties({})
        self.original_props = {}


    def on_previous_object(self):
        ''' User selected an object in observations table
        '''

        # get data for object from Metadata
        md = Component.get('Metadata').get(self.props)

        # lookup using name and object type
        reference_settings = self.cats.lookup(md.get('Name', ''), OT=md.get('OT', ''))

        # use ref settings if available, else Name/OT from metadata
        settings = md if reference_settings is None else reference_settings

        # update all properties which causes screen updates
        self.update_properties(settings)

        # store original props so as to be able to spot changes
        self.original_props = settings

        for p, val in settings.items():
            logger.info(f'{p} = {val}')


    def update_properties(self, props, update_name=True, update_OT=True):
        ''' Fill in fields with properties
        '''

        if props is None:
            return

        self.updating = True
        for p in set(self.props):
            setattr(self, p, prop_to_str(p, props.get(p, '')))
        if update_name:
            self.dso_panel.name_field.text = self.Name
        if update_OT:
            self.dso_panel.ot_field.text = self.OT
        self.updating = False
 

    def new_DSO_name(self, settings):
        ''' Called from ObservingList when user selects a name from the DSO table. 
        '''
        
        # if we already have a name and a non-empty stack, ask user to confirm
        new_name = settings.get('Name','')
        if not Component.get('Stacker').is_empty() and self.Name and self.Name != new_name:
            change = f'from {self.Name} to {new_name}'
            self.dialog = MDDialog(
                auto_dismiss=False,
                text="Are you sure you wish to change the name of the current DSO\n" + change,
                buttons=[
                    MDFlatButton(text="YES", 
                        on_press=partial(self.change_name, settings)),
                    MDFlatButton(
                        text="CANCEL", 
                        on_press=self._cancel)
                ],
            )
            self.dialog.open()
        else:
            self.change_name(settings)


    def change_name(self, settings, *args):

        if hasattr(self, 'dialog'):
            self.dialog.dismiss()

        # update properties
        self.update_properties(self.cats.lookup(
            settings.get('Name', ''), OT=settings.get('OT', '')))

        self.check_for_change()


    def _cancel(self, *args):
        self.dialog.dismiss()


    def Name_changed(self, val, *args):
        ''' Name is changed as soon as the text is altered to allow lookup
        '''

        if self.updating:
            return

        matches = self.cats.lookup(val)
        if matches is None:
            self.update_properties({}, update_name=False)
        else:
            logger.debug(f'Found match: {matches}')
            self.update_properties(matches)
        self.Name = val

        # not clear why this is commented out...
        self.check_for_change()


    def OT_changed(self, widget):
        ''' Allow any object type of 1-3 chars
        '''

        if self.updating:
            return

        widget.invalid = False
        ot = widget.text.upper()
        widget.invalid = len(ot) > 3
        if not widget.invalid:
            self.OT = ot
            props = self.cats.lookup(self.Name, OT=ot)
            if props is not None:
                self.update_properties(props)
        self.check_for_change()


    def prop_changed(self, widget, prop, update=True):
        if self.updating:
            return
        self.updating = True
        widget.invalid = False
        val = widget.text.strip()
        if val == '':
            widget.text = '='
            widget.text = ''
            setattr(self, prop, val)
        else:
            valid_prop = validated_prop(val, prop)
            widget.invalid = valid_prop is None
            if valid_prop is not None:
                if update:
                    setattr(self, prop, valid_prop)
        self.check_for_change()
        self.updating = False


    def check_for_change(self):
        ''' Compare original with current properties
        '''

        orig = {p: prop_to_str(p, self.original_props.get(p, '')) for p in  self.props}
        now = {p: prop_to_str(p, getattr(self, p)) for p in self.props}
        self.changed = 'DSO properties changed' if orig != now else ''
        if orig != now:
            logger.trace(f'orig {orig}  now {now}')
        return orig != now
 

    def on_save_object(self):
        ''' Update metadata and add any new or changed DSO to user catalogue
        '''

        props = {
            'Name': self.Name.strip(),
            'OT': self.OT.strip(),
            'RA': RA.parse(self.RA),
            'Dec': Dec.parse(self.Dec),
            'Con': self.Con.strip(),
            'Diam': str_to_arcmin(self.Diam),
            'Mag': str_to_float(self.Mag),
            'Other': self.Other.strip()
            }

        logger.trace('Checking for change')

        # update metadata
        if props['Name'] == '':
            props['Name'] = 'anon'
        Component.get('Metadata').set(props)

        known_props = self.cats.lookup(props['Name'], OT=props['OT'])

        # existing object and not changed => do nothing
        if known_props is not None:
            orig = {p: prop_to_str(p, known_props.get(p, '')) for p in self.props}
            now = {p: prop_to_str(p, getattr(self, p)) for p in self.props}
            if orig == now:
                logger.trace('known object and has not changed')
                return
 
        ''' add to user DSO catalogue if new or update if modified and has 
            well-formed name/OT/RA/Dec; also update DSO table (observing list)
        ''' 
        if props['Name'] and props['OT'] and props['RA'] is not None and props['Dec'] is not None:
            logger.trace('New or modified object and has Name/OT/RA/Dec so updating user catalogue')
            self.cats.update_user_object(props)
 

    def current_object_coordinates(self):
        ''' Called by platesolver
        '''
        if self.RA and self.Dec:
            return (float(RA(self.RA)), float(Dec(self.Dec)))
        return None, None
